#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import shutil
import tempfile
import unittest

import numpy as np

from mleap.sklearn.preprocessing.data import LabelEncoder
from mleap.sklearn.preprocessing.data import OneHotEncoder


class TestOneHotEncoder(unittest.TestCase):
    def setUp(self):
        labels = ['a', 'b', 'c', 'a', 'b', 'b']
        self.le = LabelEncoder(input_features=['label'], output_features='label_le_encoded')
        self.oh_data = self.le.fit_transform(labels).reshape(-1, 1)
        self.tmp_dir = tempfile.mkdtemp(prefix="mleap.python.tests")

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_one_hot_encoder_serialization_fails_on_multiple_feature_columns(self):
        self.oh_data = np.hstack((self.oh_data, self.oh_data))  # make two feature columns

        ohe = OneHotEncoder(handle_unknown='error')
        ohe.mlinit(prior_tf=self.le, output_features='{}_one_hot_encoded'.format(self.le.output_features))
        ohe.fit(self.oh_data)

        with self.assertRaises(NotImplementedError):
            ohe.serialize_to_bundle(self.tmp_dir, ohe.name)

    def test_one_hot_encoder_serialization_fails_on_an_invalid_category_range(self):
        self.oh_data[2][0] = 3  # make invalid category range

        ohe = OneHotEncoder(handle_unknown='error')
        ohe.mlinit(prior_tf=self.le, output_features='{}_one_hot_encoded'.format(self.le.output_features))
        ohe.fit(self.oh_data)

        with self.assertRaises(ValueError):
            ohe.serialize_to_bundle(self.tmp_dir, ohe.name)

    def test_one_hot_encoder_serialization_fails_when_using_the_drop_param(self):
        ohe = OneHotEncoder(handle_unknown='error', drop='first')  # try to use `drop` parameter
        ohe.mlinit(prior_tf=self.le, output_features='{}_one_hot_encoded'.format(self.le.output_features))
        ohe.fit(self.oh_data)

        with self.assertRaises(NotImplementedError):
            ohe.serialize_to_bundle(self.tmp_dir, ohe.name)

    def test_one_hot_encoder_serialization_fails_when_using_the_dtype_param(self):
        ohe = OneHotEncoder(handle_unknown='error', dtype=int)  # try to use `dtype` parameter
        ohe.mlinit(prior_tf=self.le, output_features='{}_one_hot_encoded'.format(self.le.output_features))
        ohe.fit(self.oh_data)

        with self.assertRaises(NotImplementedError):
            ohe.serialize_to_bundle(self.tmp_dir, ohe.name)

    def test_one_hot_encoder_serialization_succeeds_when_handle_unknown_is_set_to_error(self):
        ohe = OneHotEncoder(handle_unknown='error')
        ohe.mlinit(prior_tf=self.le, output_features='{}_one_hot_encoded'.format(self.le.output_features))
        ohe.fit(self.oh_data)

        ohe.serialize_to_bundle(self.tmp_dir, ohe.name)
        with open("{}/{}.node/model.json".format(self.tmp_dir, ohe.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual('one_hot_encoder', model['op'])
        self.assertEqual(3, model['attributes']['size']['long'])
        self.assertEqual('error', model['attributes']['handle_invalid']['string'])
        self.assertEqual(False, model['attributes']['drop_last']['boolean'])

    def test_one_hot_encoder_deserialization_succeeds_when_handle_unknown_is_set_to_error(self):
        ohe = OneHotEncoder(handle_unknown='error')
        ohe.mlinit(prior_tf=self.le, output_features='{}_one_hot_encoded'.format(self.le.output_features))
        ohe.fit(self.oh_data)

        ohe.serialize_to_bundle(self.tmp_dir, ohe.name)

        node_name = "{}.node".format(ohe.name)
        ohe_ds = OneHotEncoder()
        ohe_ds.deserialize_from_bundle(self.tmp_dir, node_name)

        self.oh_data[2][0] = 3  # Add an unknown category

        with self.assertRaises(ValueError):
            ohe_ds.transform(self.oh_data)

    def test_one_hot_encoder_serialization_succeeds_when_handle_unknown_is_set_to_ignore(self):
        labels = ['a', 'b', 'c', 'a', 'b', 'b']

        le = LabelEncoder(input_features=['label'], output_features='label_le_encoded')
        oh_data = le.fit_transform(labels).reshape(-1, 1)

        ohe = OneHotEncoder(handle_unknown='ignore')
        ohe.mlinit(prior_tf=le, output_features='{}_one_hot_encoded'.format(le.output_features))
        ohe.fit(oh_data)

        ohe.serialize_to_bundle(self.tmp_dir, ohe.name)
        with open("{}/{}.node/model.json".format(self.tmp_dir, ohe.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual('one_hot_encoder', model['op'])
        self.assertEqual(3, model['attributes']['size']['long'])
        self.assertEqual('keep', model['attributes']['handle_invalid']['string'])
        self.assertEqual(True, model['attributes']['drop_last']['boolean'])

    def test_one_hot_encoder_deserialization_succeeds_when_handle_unknown_is_set_to_ignore(self):
        ohe = OneHotEncoder(handle_unknown='ignore')
        ohe.mlinit(prior_tf=self.le, output_features='{}_one_hot_encoded'.format(self.le.output_features))
        ohe.fit(self.oh_data)

        ohe.serialize_to_bundle(self.tmp_dir, ohe.name)

        node_name = "{}.node".format(ohe.name)
        ohe_ds = OneHotEncoder()
        ohe_ds.deserialize_from_bundle(self.tmp_dir, node_name)

        self.oh_data[2][0] = 3  # Add an unknown category

        expected = ohe.transform(self.oh_data).todense()
        actual = ohe_ds.transform(self.oh_data)

        np.testing.assert_array_equal(expected, actual)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from sklearn.tree import DecisionTreeClassifier
from sklearn.tree import DecisionTreeRegressor
from mleap.bundle.serialize import MLeapSerializer
from sklearn.tree import _tree
import json
import uuid


def mleap_init(self, input_features, prediction_column, feature_names):
    self.input_features = input_features
    self.prediction_column = prediction_column
    self.feature_names = feature_names
    self.name = "{}_{}".format(self.op, uuid.uuid1())


def serialize_to_bundle(self, path, model_name, serialize_node=True):
    serializer = SimpleSerializer()
    return serializer.serialize_to_bundle(self, path, model_name, serialize_node=serialize_node)


setattr(DecisionTreeRegressor, 'op', 'decision_tree_regression')
setattr(DecisionTreeRegressor, 'mlinit', mleap_init)
setattr(DecisionTreeRegressor, 'serialize_to_bundle', serialize_to_bundle)
setattr(DecisionTreeRegressor, 'serializable', True)

setattr(DecisionTreeClassifier, 'op', 'decision_tree_classifier')
setattr(DecisionTreeClassifier, 'mlinit', mleap_init)
setattr(DecisionTreeClassifier, 'serialize_to_bundle', serialize_to_bundle)
setattr(DecisionTreeClassifier, 'serializable', True)


class SimpleSerializer(MLeapSerializer):
    def __init__(self):
        super(SimpleSerializer, self).__init__()

    @staticmethod
    def serialize_tree(tree, feature_names, outfile):
        """
        :type feature_names: list
        :type tree: sklearn.tree.tree.BaseDecisionTree
        :param tree: sklearn.tree.tree
        :param feature_names:
        :return:
        """

        tree_ = tree.tree_
        feature_name = [feature_names[i] if i != _tree.TREE_UNDEFINED else 'n/a' for i in tree_.feature]

        def traverse(node, depth, outfile):
            if tree_.feature[node] != _tree.TREE_UNDEFINED:
                name = feature_name[node]
                threshold = tree_.threshold[node]

                # Define internal node for serialization
                internal_node = {
                    'type': 'internal',
                    'split': {
                        'type': 'continuous',
                        'featureIndex': feature_names.index(name),
                        'threshold': threshold
                    }
                }

                # Serialize the internal Node
                json.dump(internal_node, outfile)
                outfile.write('\n')

                # Traverse Left
                traverse(tree_.children_left[node], depth + 1, outfile)

                # Traverse Rigiht
                traverse(tree_.children_right[node], depth + 1, outfile)
            else:
                leaf_node = {
                    'type': 'leaf',
                    'values': tree_.value[node].tolist()[0]
                }

                # Serialize the leaf node
                json.dump(leaf_node, outfile)
                outfile.write('\n')

        traverse(0, 1, outfile)

    def serialize_to_bundle(self, transformer, path, model_name, serialize_node=True):
        """
        :type transformer: sklearn.tree.tree.BaseDecisionTree
        :type path: str
        :type model_name: str
        :type serialize_node: bool
        :param transformer:
        :param path:
        :param model_name:
        :return:
        """

        # Define attributes
        attributes = list()
        attributes.append(('num_features', transformer.n_features_))
        if isinstance(transformer, DecisionTreeClassifier):
            attributes.append(('num_classes', int(transformer.n_classes_)))

        inputs = []
        outputs = []
        if serialize_node:
            # define node inputs and outputs
            inputs = [{
                      "name": transformer.input_features,
                      "port": "features"
                    }]

            outputs = [{
                      "name": transformer.prediction_column,
                      "port": "prediction"
                    }]

        self.serialize(transformer, path, model_name, attributes, inputs, outputs, node=serialize_node)

        # Serialize tree.json
        tree_path = "{}/{}.node/tree.json".format(path, model_name)
        if not serialize_node:
            tree_path = "{}/{}/tree.json".format(path, model_name)
        with open(tree_path, 'w') as outfile:
            self.serialize_tree(transformer, transformer.feature_names, outfile)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from gensim.models import Word2Vec
from mleap.bundle.serialize import MLeapSerializer
import uuid
import numpy as np


def serialize_to_bundle(self, path, model_name):
    serializer = SimpleSparkSerializer()
    return serializer.serialize_to_bundle(self, path, model_name)


def sent2vec(self, words):
    serializer = SimpleSparkSerializer()
    return serializer.sent2vec(words, self)


def mleap_init(self, input_features, prediction_column):
    self.input_features = input_features
    self.prediction_column = prediction_column
    self.name = "{}_{}".format(self.op, uuid.uuid4())

setattr(Word2Vec, 'op', 'word2vec')
setattr(Word2Vec, 'mlinit', mleap_init)
setattr(Word2Vec, 'serialize_to_bundle', serialize_to_bundle)
setattr(Word2Vec, 'serializable', True)
setattr(Word2Vec, 'sent2vec', sent2vec)


class SimpleSparkSerializer(MLeapSerializer):
    def __init__(self):
        super(SimpleSparkSerializer, self).__init__()

    @staticmethod
    def set_prediction_column(transformer, prediction_column):
        transformer.prediction_column = prediction_column

    @staticmethod
    def set_input_features(transformer, input_features):
        transformer.input_features = input_features

    def serialize_to_bundle(self, transformer, path, model_name):

        # compile tuples of model attributes to serialize
        attributes = list()
        attributes.append(('words', transformer.wv.index_to_key))

        # indices = [np.float64(x) for x in list(range(len(transformer.wv.index_to_key)))]
        word_vectors = np.array([float(y) for x in [transformer.wv.word_vec(w) for w in transformer.wv.index_to_key] for y in x])

        # attributes.append(('indices', indices))
        # Excluding indices because they are 0 - N
        attributes.append(('word_vectors', word_vectors))
        attributes.append(('kernel', 'sqrt'))

        # define node inputs and outputs
        inputs = [{
                  "name": transformer.input_features,
                  "port": "input"
                  }]

        outputs = [{
                  "name": transformer.prediction_column,
                  "port": "output"
                   }]

        self.serialize(transformer, path, model_name, attributes, inputs, outputs)

    def sent2vec(self, words, transformer):
        """
        Used with sqrt kernel
        :param words:
        :param transformer:
        :return:
        """
        sent_vec = np.zeros(transformer.vector_size)
        numw = 0
        for w in words:
            try:
                sent_vec = np.add(sent_vec, transformer.wv[w])
                numw += 1
            except:
                continue
        return sent_vec / np.sqrt(sent_vec.dot(sent_vec))



import os
import tempfile
import unittest

from py4j.protocol import Py4JJavaError
from pyspark.ml import Pipeline
from pyspark.sql import types as t

from mleap.pyspark.feature.string_map import StringMap
from mleap.pyspark.spark_support import SimpleSparkSerializer
from tests.pyspark.lib.assertions import assert_df
from tests.pyspark.lib.spark_session import spark_session


INPUT_SCHEMA = t.StructType([t.StructField('key_col', t.StringType(), False),
                             t.StructField('extra_col', t.StringType(), False)])

OUTPUT_SCHEMA = t.StructType([t.StructField('key_col', t.StringType(), False),
                              t.StructField('extra_col', t.StringType(), False),
                              t.StructField('value_col', t.DoubleType(), False)])


class StringMapTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.spark = spark_session()

    @classmethod
    def tearDownClass(cls):
        cls.spark.stop()

    def setUp(self):
        self.input = StringMapTest.spark.createDataFrame([['a', 'b']], INPUT_SCHEMA)

    def test_map(self):
        result = StringMap(
            labels={'a': 1.0},
            inputCol='key_col',
            outputCol='value_col',
        ).transform(self.input)
        expected = StringMapTest.spark.createDataFrame([['a', 'b', 1.0]], OUTPUT_SCHEMA)
        assert_df(expected, result)

    def test_map_default_value(self):
        result = StringMap(
            labels={'z': 1.0},
            inputCol='key_col',
            outputCol='value_col',
            handleInvalid='keep',
        ).transform(self.input)
        expected = StringMapTest.spark.createDataFrame([['a', 'b', 0.0]], OUTPUT_SCHEMA)
        assert_df(expected, result)

    def test_map_custom_default_value(self):
        result = StringMap(
            labels={'z': 1.0},
            inputCol='key_col',
            outputCol='value_col',
            handleInvalid='keep',
            defaultValue=-1.0
        ).transform(self.input)
        expected = StringMapTest.spark.createDataFrame([['a', 'b', -1.0]], OUTPUT_SCHEMA)
        assert_df(expected, result)

    def test_map_missing_value_error(self):
        with self.assertRaises(Py4JJavaError) as error:
            StringMap(
                labels={'z': 1.0},
                inputCol='key_col',
                outputCol='value_col'
            ).transform(self.input).collect()
        self.assertIn('java.util.NoSuchElementException: Missing label: a', str(error.exception))

    def test_map_from_dataframe(self):
        labels_df = StringMapTest.spark.createDataFrame([['a', 1.0]], 'key_col: string, value_col: double')
        result = StringMap.from_dataframe(
            labels_df=labels_df,
            inputCol='key_col',
            outputCol='value_col'
        ).transform(self.input)
        expected = StringMapTest.spark.createDataFrame([['a', 'b', 1.0]], OUTPUT_SCHEMA)
        assert_df(expected, result)

    def test_serialize_to_bundle(self):
        string_map = StringMap(
            labels={'a': 1.0},
            inputCol='key_col',
            outputCol='value_col',
        )
        pipeline = Pipeline(stages=[string_map]).fit(self.input)
        serialization_dataset = pipeline.transform(self.input)

        jar_file_path = _serialize_to_file(pipeline, serialization_dataset)
        deserialized_pipeline = _deserialize_from_file(jar_file_path)

        result = deserialized_pipeline.transform(self.input)
        expected = StringMapTest.spark.createDataFrame([['a', 'b', 1.0]], OUTPUT_SCHEMA)
        assert_df(expected, result)

    @staticmethod
    def test_validate_handleInvalid_ok():
        StringMap(labels={}, handleInvalid='error')

    def test_validate_handleInvalid_bad(self):
        with self.assertRaises(AssertionError):
            StringMap(labels=None, inputCol=dict(), outputCol=None, handleInvalid='invalid')

    def test_validate_labels_type_fails(self):
        with self.assertRaises(AssertionError):
            StringMap(labels=None, inputCol=set(), outputCol=None)

    def test_validate_labels_key_fails(self):
        with self.assertRaises(AssertionError):
            StringMap(labels=None, inputCol={False: 0.0}, outputCol=None)

    def test_validate_labels_value_fails(self):
        with self.assertRaises(AssertionError):
            StringMap(labels=None, inputCol={'valid_key_type': 'invalid_value_type'}, outputCol=None)


def _serialize_to_file(model, df_for_serializing):
    jar_file_path = _to_jar_file_path(
        os.path.join(tempfile.mkdtemp(), 'test_serialize_to_bundle-pipeline.zip'))
    SimpleSparkSerializer().serializeToBundle(model, jar_file_path, df_for_serializing)
    return jar_file_path


def _to_jar_file_path(path):
    return "jar:file:" + path


def _deserialize_from_file(path):
    return SimpleSparkSerializer().deserializeFromBundle(path)



import six
from pyspark import keyword_only
from pyspark.ml.util import JavaMLReadable, JavaMLWritable, _jvm
from pyspark.ml.wrapper import JavaTransformer
from pyspark.ml.param.shared import HasInputCol, HasOutputCol
from pyspark.sql import DataFrame

from mleap.pyspark.py2scala import jvm_scala_object


class StringMap(JavaTransformer, HasInputCol, HasOutputCol, JavaMLReadable, JavaMLWritable):

    @keyword_only
    def __init__(self, labels={}, inputCol=None, outputCol=None, handleInvalid='error', defaultValue=0.0):
        """
        :param labels: a dict {string: double}
        :param handleInvalid: how to handle missing labels: 'error' (throw), or 'keep' (map to defaultValue)
        :param defaultValue: value to use if key is not found in labels
        """
        """
        labels must be a dict {string: double} or a spark DataFrame with columns inputCol & outputCol
        handleInvalid: 
        """
        super(StringMap, self).__init__()

        def validate_args():
            """
            validate args early to avoid failing at Py4j with some hard to interpret error message
            """
            assert handleInvalid in ['error', 'keep'], 'Invalid value for handleInvalid: {}'.format(handleInvalid)
            assert isinstance(labels, dict), 'labels must be a dict, got: {}'.format(type(labels))
            for (key, value) in labels.items():
                assert isinstance(key, six.string_types), \
                    'label keys must be a string type, got: {}'.format(type(key))
                assert isinstance(value, float), 'label values must be float, got: {}'.format(type(key))

        validate_args()

        labels_scala_map = _jvm() \
            .scala \
            .collection \
            .JavaConverters \
            .mapAsScalaMapConverter(labels) \
            .asScala() \
            .toMap(_jvm().scala.Predef.conforms())

        handle_invalid_jvm = jvm_scala_object(
            _jvm().ml.combust.mleap.core.feature,
            f"HandleInvalid${handleInvalid.capitalize()}$",
        )

        string_map_model = _jvm().ml.combust.mleap.core.feature.StringMapModel(
            labels_scala_map,
            handle_invalid_jvm,
            defaultValue,
        )

        self._java_obj = self._new_java_obj(
            "org.apache.spark.ml.mleap.feature.StringMap",
            self.uid,
            string_map_model
        )

        self._setDefault()
        self.setParams(inputCol=inputCol, outputCol=outputCol)

    @keyword_only
    def setParams(self, inputCol=None, outputCol=None):
        """
        setParams(self, inputCol=None, outputCol=None)
        Sets params for this StringMap.
        """
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    def setInputCol(self, value):
        """
        Sets the value of :py:attr:`inputCol`.
        """
        return self._set(inputCol=value)

    def setOutputCol(self, value):
        """
        Sets the value of :py:attr:`outputCol`.
        """
        return self._set(outputCol=value)

    @classmethod
    def from_dataframe(cls, labels_df, inputCol, outputCol, handleInvalid='error', defaultValue=0.0):
        """
        :param labels_df: a spark DataFrame whose columns include inputCol:string & outputCol:double.
        See StringMap() for other params.
        """
        assert isinstance(labels_df, DataFrame), 'labels must be a DataFrame, got: {}'.format(type(labels_df))
        labels_dict = {
            row[0]: float(row[1])
            for row in
            labels_df.select([inputCol, outputCol]).collect()
        }

        return cls(
            labels=labels_dict,
            inputCol=inputCol,
            outputCol=outputCol,
            handleInvalid=handleInvalid,
            defaultValue=defaultValue
        )




#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from sklearn.pipeline import Pipeline
from mleap import __version__
import os
import json
import shutil
import uuid
import zipfile
import datetime

def serialize_to_bundle(self, path, model_name, init=False):
    serializer = SimpleSerializer()
    serializer.serialize_to_bundle(self, path, model_name, init)


def deserialize_from_bundle(self, path):
    serializer = SimpleSerializer()
    return serializer.deserialize_from_bundle(path)


def mleap_init(self):
    self.name = "{}_{}".format(self.op, uuid.uuid1())


setattr(Pipeline, 'serialize_to_bundle', serialize_to_bundle)
setattr(Pipeline, 'deserialize_from_bundle', deserialize_from_bundle)
setattr(Pipeline, 'op', 'pipeline')
setattr(Pipeline, 'mlinit', mleap_init)
setattr(Pipeline, 'serializable', True)


class SimpleSerializer(object):
    def __init__(self):
        super(SimpleSerializer, self).__init__()

    def serialize_to_bundle(self, transformer, path, model_name, init=False):

        model_dir = path
        if init:
            # If bundle path already exists, delte it and create a clean directory
            if os.path.exists("{}/{}".format(path, model_name)):
                shutil.rmtree("{}/{}".format(path, model_name))

            # make pipeline directory
            model_dir = "{}/{}".format(path, model_name)
            os.mkdir(model_dir)

            # make pipeline root directory
            root_dir = "{}/root".format(model_dir)
            os.mkdir(root_dir)

            # Write Pipeline Bundle file
            with open("{}/{}".format(model_dir, 'bundle.json'), 'w') as outfile:
                json.dump(self.get_bundle(transformer), outfile, indent=3)

            model_dir = root_dir

            # Write the model and node files
            with open("{}/{}".format(model_dir, 'model.json'), 'w') as outfile:
                json.dump(self.get_model(transformer), outfile, indent=3)

            with open("{}/{}".format(model_dir, 'node.json'), 'w') as outfile:
                json.dump(self.get_node(transformer), outfile, indent=3)

        else:
            # Write model file
            with open("{}/{}".format(model_dir, 'model.json'), 'w') as outfile:
                json.dump(self.get_model(transformer), outfile, indent=3)

            # Write node file
            with open("{}/{}".format(model_dir, 'node.json'), 'w') as outfile:
                json.dump(self.get_node(transformer), outfile, indent=3)

        for step in [x[1] for x in transformer.steps if hasattr(x[1], 'serialize_to_bundle')]:
            name = step.name

            if step.op == 'pipeline':
                # Create the node directory
                bundle_dir = "{}/{}.node".format(model_dir, name)
                os.mkdir(bundle_dir)

                # Write model file
                with open("{}/{}".format(bundle_dir, 'model.json'), 'w') as outfile:
                    json.dump(self.get_model(step), outfile, indent=3)

                # Write node file
                with open("{}/{}".format(bundle_dir, 'node.json'), 'w') as outfile:
                    json.dump(self.get_node(step), outfile, indent=3)

                for step_i in [x[1] for x in step.steps]:
                    step_i.serialize_to_bundle(bundle_dir, step_i.name)

            elif step.op == 'feature_union':
                step.serialize_to_bundle(model_dir, step.name)
            else:
                step.serialize_to_bundle(model_dir, step.name)

            if isinstance(step, list):
                pass

        if init:
            zip_pipeline(path, model_name)

    def deserialize_from_bundle(self, path):
        return NotImplementedError

    @staticmethod
    def get_bundle(transformer):
        js = {
          "name": transformer.name,
          "format": "json",
          "version": __version__,
          "timestamp": datetime.datetime.now().isoformat(),
          "uid": "{}".format(uuid.uuid4())
        }
        return js

    @staticmethod
    def get_node(transformer):
        js = {
          "name": transformer.name,
          "shape": {
            "inputs": [],
            "outputs": []
          }
        }
        return js

    def get_model(self, transformer):
        js = {
          "op": transformer.op,
            "attributes": {
                "nodes": {
                    "type": "list",
                    "string": self._extract_nodes(transformer.steps)
                }
            }
        }
        return js

    @staticmethod
    def _extract_nodes(steps):
        pipeline_steps = []
        for name, step in steps:
            if step.op == 'feature_union':
                union_steps = [x[1].name for x in step.transformer_list if hasattr(x[1], 'serialize_to_bundle') and x[1].serializable]
                pipeline_steps += union_steps
            elif hasattr(step, 'serialize_to_bundle') and step.serializable:
                pipeline_steps.append(step.name)
        return pipeline_steps


def zip_pipeline(path, name):
    zip_file = zipfile.ZipFile("{}/{}.zip".format(path, name), 'w', zipfile.ZIP_DEFLATED)
    abs_src = os.path.abspath("{}/{}".format(path, name))
    for root, dirs, files in os.walk("{}/{}".format(path, name)):
        for file in files:
            absname = os.path.abspath(os.path.join(root, file))
            arcname = absname[len(abs_src) + 1:]

            zip_file.write(os.path.join(root, file), arcname)
    zip_file.close()


#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import uuid

import numpy as np
import pandas as pd
from mleap.sklearn.preprocessing.data import ImputerSerializer, FeatureExtractor
from sklearn.impute import SimpleImputer as SKLearnImputer
from sklearn.base import BaseEstimator, TransformerMixin


class DefineEstimator(BaseEstimator, TransformerMixin):
    def __init__(self, transformer):
        """
        Selects all but the last column of a matrix and passes it as the X variable into the transformer,
        and the last column as the y variable.

        This transformer is useful when we need to run a transformer or a series of transformers on the y-variable
        :type transformer: BaseEstimator
        :param transformer: Estimator (linear regression, random forest, etc)
        :return: Estimator
        """
        self.op = 'define_estimator'
        self.name = "{}_{}".format(self.op, uuid.uuid4())
        self.transformer = transformer
        self.serializable=False

    def fit(self, X, y, sample_weight=None):
        return self.transformer.fit(X[:, :-1], X[:, -1:])

    def transform(self, X):
        return self.transformer.predict(X[:,:-1])

    def fit_transform(self, X, y=None, **fit_params):
        return self.transformer.fit_transform(X[:,:-1], X[:,-1:])

    def predict(self, X):
        return self.transformer.predict(X[:,:-1])


"""
Wrapper around the sklearn Imputer so that it can be used in pipelines. 
Delegates fit() and transform() methods to the sklearn transformer and uses the ImputerSerializer to serialize to bundle.

Instead of putting a FeatureExtractor ahead of the Imputer, we add the equivalent of FeatureExtractor's transform() method
in the fit() and transform() methods.

This is because the Imputer both in Spark and MLeap operates on a scalar value and if we were to add a feature extractor in
front of it, then it would serialize as operating on a tensor and thus, fail at scoring time. 
"""
class Imputer(SKLearnImputer):

    def __init__(self, input_features, output_features,
                 missing_values=np.nan, strategy="mean",
                 fill_value=None, verbose=0, copy=True, add_indicator=False):
        self.name = "{}_{}".format(self.op, uuid.uuid1())
        self.input_features = input_features
        self.output_features = output_features
        self.input_shapes = {'data_shape': [{'shape': 'scalar'}]}
        self.feature_extractor = FeatureExtractor(input_scalars=[input_features],
                                                  output_vector='extracted_' + output_features,
                                                  output_vector_items=[output_features])
        SKLearnImputer.__init__(
            self,
            missing_values=missing_values,
            strategy=strategy,
            fill_value=fill_value,
            verbose=verbose,
            copy=copy,
            add_indicator=add_indicator,
           )

    def fit(self, X, y=None):
        super(Imputer, self).fit(self.feature_extractor.transform(X))
        return self

    def transform(self, X):
        return pd.DataFrame(super(Imputer, self).transform(self.feature_extractor.transform(X)))

    def serialize_to_bundle(self, path, model_name):
        ImputerSerializer().serialize_to_bundle(self, path, model_name)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import shutil
import json
import os
import tempfile
import unittest
import uuid

from mleap.gensim.word2vec import Word2Vec


class TransformerTests(unittest.TestCase):

    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_word2vec(self):

        sentences4word2vec_ = [
            ['call', 'me', 'tomorrow'],
            ['give', 'me', 'a', 'call', 'in',' the', 'after', 'noon'],
            ['when', 'can', 'i', 'call'],
            ['when', 'is', 'the', 'best', 'time', 'to', 'call'],
            ['call', 'me', 'tomorrow', 'after', 'noon'],
            ['i', 'would', 'like', 'a', 'call', 'tomorrow'],
            ['do', 'not', 'call', 'until', 'tomorrow'],
            ['best', 'time', 'is', 'tomorrow', 'after', 'noon'],
            ['call', 'tomorrow', 'after', 'lunch'],
            ['call', 'after', 'lunch', 'time'],
            ['make', 'the', 'call', 'tomorrow'],
            ['make', 'the', 'call', 'tomorrow', 'after', 'noon'],
            ['make', 'the' 'call', 'after', 'lunch', 'time']
        ]

        size_ = 5
        window_ = 2

        model_ = Word2Vec(sentences4word2vec_, min_count=2, vector_size=size_, window=window_)
        model_.mlinit(input_features=['input'], prediction_column = 'sentence_vector')

        model_.serialize_to_bundle(self.tmp_dir, model_.name)

        res = model_.sent2vec(['call', 'me', 'on', 'my', 'cell', 'phone'])


        with open('{}/{}.node/node.json'.format(self.tmp_dir, model_.name)) as node_json:
            node = json.load(node_json)

        with open('{}/{}.node/model.json'.format(self.tmp_dir, model_.name)) as model_json:
            model = json.load(model_json)


        self.assertEqual(5, res.size)
        self.assertEqual(node['shape']['inputs'][0]['name'], ['input'])
        self.assertEqual(node['shape']['outputs'][0]['name'], 'sentence_vector')
        self.assertEqual(model['op'], 'word2vec')



import sys

import pyspark.ml

import mleap.pyspark
import mleap.pyspark.feature
from mleap.pyspark.feature.string_map import StringMap
from mleap.pyspark.feature.math_binary import MathBinary
from mleap.pyspark.feature.math_unary import MathUnary

sys.modules['pyspark.ml.mleap'] = mleap
sys.modules['pyspark.ml.mleap.pyspark'] = sys.modules['mleap.pyspark']
sys.modules['pyspark.ml.mleap.feature'] = sys.modules['mleap.pyspark.feature']

sys.modules['pyspark.ml'].mleap = mleap
sys.modules['pyspark.ml'].mleap.feature = sys.modules['mleap.pyspark.feature']
sys.modules['pyspark.ml'].mleap.feature.StringMap = StringMap
sys.modules['pyspark.ml'].mleap.feature.MathUnary = MathUnary
sys.modules['pyspark.ml'].mleap.feature.MathBinary = MathBinary



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from sklearn.ensemble.forest import RandomForestRegressor
from sklearn.ensemble.forest import RandomForestClassifier
from mleap.bundle.serialize import MLeapSerializer
from mleap.bundle.serialize import Vector
import mleap.sklearn.tree.tree
import uuid


def mleap_init(self, input_features, prediction_column, feature_names):
    self.input_features = input_features
    self.prediction_column = prediction_column
    self.feature_names = feature_names
    self.name = "{}_{}".format(self.op, uuid.uuid1())


def serialize_to_bundle(self, path, model_name):
    serializer = SimpleSerializer()
    return serializer.serialize_to_bundle(self, path, model_name)


setattr(RandomForestRegressor, 'op', 'random_forest_regression')
setattr(RandomForestRegressor, 'mlinit', mleap_init)
setattr(RandomForestRegressor, 'serialize_to_bundle', serialize_to_bundle)
setattr(RandomForestRegressor, 'serializable', True)

setattr(RandomForestClassifier, 'op', 'random_forest_classifier')
setattr(RandomForestClassifier, 'mlinit', mleap_init)
setattr(RandomForestClassifier, 'serialize_to_bundle', serialize_to_bundle)
setattr(RandomForestClassifier, 'serializable', True)


class SimpleSerializer(MLeapSerializer):
    def __init__(self):
        super(SimpleSerializer, self).__init__()

    def serialize_to_bundle(self, transformer, path, model):
        """
        :param transformer: Random Forest Regressor or Classifier
        :param path: Root path where to serialize the model
        :param model: Name of the model to be serialized
        :type transformer: sklearn.ensemble.forest.BaseForest
        :type path: str
        :type model: str
        :return: None
        """

        # Define Node Inputs and Outputs
        inputs = [{
                  "name": transformer.input_features,
                  "port": "features"
                }]

        outputs = list()
        outputs.append({
                  "name": transformer.prediction_column,
                  "port": "prediction"
                })

        outputs.append({
              "name": "raw_prediction",
              "port": "raw_prediction"
             })

        outputs.append({
              "name": "probability",
              "port": "probability"
            })

        # compile tuples of model attributes to serialize
        tree_weights = Vector([1.0 for x in range(0, len(transformer.estimators_))])
        attributes = list()
        attributes.append(('num_features', transformer.n_features_))
        attributes.append(('tree_weights', tree_weights))
        attributes.append(('trees', ["tree{}".format(x) for x in range(0, len(transformer.estimators_))]))
        if isinstance(transformer, RandomForestClassifier):
            attributes.append(('num_classes', transformer.n_classes_)) # TODO: get number of classes from the transformer

        self.serialize(transformer, path, model, attributes, inputs, outputs)

        rf_path = "{}/{}.node".format(path, model)

        estimators = transformer.estimators_

        i = 0
        for estimator in estimators:
            estimator.mlinit(input_features = transformer.input_features, prediction_column = transformer.prediction_column, feature_names=transformer.feature_names)
            model_name = "tree{}".format(i)
            estimator.serialize_to_bundle(rf_path, model_name, serialize_node=False)

            i += 1



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer
from mleap.sklearn.pipeline import SimpleSerializer as PipelineSerializer
from mleap.bundle.serialize import MLeapSerializer
import uuid


class ops(object):
    def __init__(self):
        self.COUNT_VECTORIZER = 'tokenizer'
        self.TF_IDF_VECTORIZER = 'pipeline'


ops = ops()


def serialize_to_bundle(self, path, model_name):
    serializer = SimpleSerializer()
    return serializer.serialize_to_bundle(self, path, model_name)


def mleap_init(self, input_features, prediction_column):
    name = "{}_{}"

    if isinstance(self, TfidfVectorizer):
        name = "tfidf_{}_{}"

    self.name = name.format(self.op, uuid.uuid4())
    self.input_features = input_features
    self.prediction_column = prediction_column


setattr(TfidfVectorizer, 'op', ops.TF_IDF_VECTORIZER)
setattr(TfidfVectorizer, 'mlinit', mleap_init)
setattr(TfidfVectorizer, 'serialize_to_bundle', serialize_to_bundle)
setattr(TfidfVectorizer, 'serializable', True)

setattr(CountVectorizer, 'op', ops.COUNT_VECTORIZER)
setattr(CountVectorizer, 'mlinit', mleap_init)
setattr(CountVectorizer, 'serialize_to_bundle', serialize_to_bundle)
setattr(CountVectorizer, 'serializable', True)


class SimpleSerializer(object):
    def __init__(self):
        super(SimpleSerializer, self).__init__()

    @staticmethod
    def _choose_serializer(transformer):
        serializer = None

        if transformer.op == ops.COUNT_VECTORIZER:
            serializer = CountVectorizerSerializer()
        elif transformer.op == ops.TF_IDF_VECTORIZER:
            serializer = TfidfVectorizerSerializer()

        return serializer

    @staticmethod
    def set_prediction_column(transformer, prediction_column):
        transformer.prediction_column = prediction_column

    @staticmethod
    def set_input_features(transformer, input_features):
        transformer.input_features = input_features

    def serialize_to_bundle(self, transformer, path, model_name):
        serializer = self._choose_serializer(transformer)
        serializer.serialize_to_bundle(transformer, path, model_name)


class CountVectorizerSerializer(MLeapSerializer):
    def __init__(self):
        super(CountVectorizerSerializer, self).__init__()

    def serialize_to_bundle(self, transformer, path, model_name):

        # compile tuples of model attributes to serialize
        attributes = None

        # define node inputs and outputs
        inputs = [{
                  "name": transformer.input_features,
                  "port": "input"
                  }]

        outputs = [{
                  "name": transformer.prediction_column,
                  "port": "output"
                   }]

        self.serialize(transformer, path, model_name, attributes, inputs, outputs)


class TfidfVectorizerSerializer(MLeapSerializer):
    pipeline_serializer = PipelineSerializer()

    def __init__(self):
        super(TfidfVectorizerSerializer, self).__init__()

    @staticmethod
    def set_prediction_column(transformer, prediction_column):
        transformer.prediction_column = prediction_column

    @staticmethod
    def set_input_features(transformer, input_features):
        transformer.input_features = input_features

    def serialize_to_bundle(self, transformer, path, model_name):
        num_features = transformer.idf_.shape[0]
        vocabulary = [None] * num_features

        for term, termidx in transformer.vocabulary_.items():
            vocabulary[termidx] = str(term)

        tf_attributes = [
            ('vocabulary', vocabulary),
            ('binary', transformer.binary),
            # no concept of 'min_tf' in scikit's TfidfVectorizer
            ('min_tf', 0)
        ]
        tf_inputs = [{
            "name": transformer.input_features,
            "port": "input"
        }]
        tf_outputs = [{
            "name": "token_counts",
            "port": "output"
        }]
        tf_step = TfidfStep(transformer, 'count_vectorizer', tf_attributes, tf_inputs, tf_outputs)

        idf_attributes = [
            ('idf', transformer.idf_.tolist())
        ]
        idf_inputs = [{
            "name": tf_outputs[0]['name'],
            "port": "input"
        }]
        idf_outputs = [{
            "name": transformer.prediction_column,
            "port": "output"
        }]
        idf_step = TfidfStep(transformer, 'idf', idf_attributes, idf_inputs, idf_outputs)

        pipeline_step = TfidfStep(transformer, 'pipeline', None, None, None)
        pipeline_step.steps = [('tf', tf_step), ('idf', idf_step)]
        self.pipeline_serializer.serialize_to_bundle(pipeline_step, path, model_name, True)


class TfidfStep(MLeapSerializer):
    def __init__(self, original_transformer, op, attributes, inputs, outputs):
        super(TfidfStep, self).__init__()

        self.op = op
        self.attributes = attributes
        self.inputs = inputs
        self.outputs = outputs
        self.name = "{}_{}".format(original_transformer.name, op)
        self.serializable = True

    def serialize_to_bundle(self, path, model_name):
        self.serialize(self, path, model_name, self.attributes, self.inputs, self.outputs)



import json
import os
import shutil
import tempfile
import unittest
import uuid

from mleap.sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer


class TransformerTests(unittest.TestCase):
    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()

        self.docs = ['test']

        self.tfidf = TfidfVectorizer(binary=False)
        self.tfidf.mlinit(input_features='some_text', prediction_column='features')
        self.tfidf.fit(self.docs)

        self.assertEqual(self.tfidf.vocabulary_, {u'test': 0})
        self.assertEqual(self.tfidf.idf_, [1.0])

        self.tfidf.serialize_to_bundle(self.tmp_dir, self.tfidf.name)
        self.pipe_model = self.load_pipeline_model()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_countvectorizer_serializer(self):
        count = CountVectorizer()
        count.mlinit(input_features='input', prediction_column='pred')
        count.fit(self.docs)
        count.serialize_to_bundle(self.tmp_dir, count.name)

        with open('{}/{}.node/node.json'.format(self.tmp_dir, count.name)) as node_json:
            node = json.load(node_json)

        with open('{}/{}.node/model.json'.format(self.tmp_dir, count.name)) as model_json:
            model = json.load(model_json)

        self.assertEqual(node['shape']['inputs'][0]['name'], 'input')
        self.assertEqual(node['shape']['outputs'][0]['name'], 'pred')
        self.assertEqual(model['op'], 'tokenizer')

    def test_tfidf_vectorizer_serializer_pipeline_part(self):
        self.assertEqual(self.pipe_model['op'], 'pipeline')

    def test_tfidf_vectorizer_serializer_tf_part(self):
        nodes = self.pipe_model['attributes']['nodes']['string']
        tf_node_name = nodes[0]

        with open('{}/{}/root/{}.node/model.json'.format(self.tmp_dir, self.tfidf.name, tf_node_name)) as tf_model_json:
            tf_model = json.load(tf_model_json)

        self.assertEqual(tf_model['attributes']['binary']['boolean'], False)
        self.assertEqual(tf_model['attributes']['vocabulary']['string'], self.docs)
        self.assertEqual(tf_model['op'], 'count_vectorizer')

        with open('{}/{}/root/{}.node/node.json'.format(self.tmp_dir, self.tfidf.name, tf_node_name)) as tf_node_json:
            tf_node = json.load(tf_node_json)

        self.assertEqual(tf_node['shape']['inputs'][0]['name'], 'some_text')
        self.assertEqual(tf_node['shape']['outputs'][0]['name'], 'token_counts')

    def test_tfidf_vectorizer_serializer_idf_part(self):
        nodes = self.pipe_model['attributes']['nodes']['string']
        idf_node_name = nodes[1]

        with open('{}/{}/root/{}.node/node.json'.format(self.tmp_dir, self.tfidf.name, idf_node_name)) as idf_node_json:
            idf_node = json.load(idf_node_json)

        self.assertEqual(idf_node['shape']['inputs'][0]['name'], 'token_counts')
        self.assertEqual(idf_node['shape']['outputs'][0]['name'], 'features')

        idf_model_file = '{}/{}/root/{}.node/model.json'.format(self.tmp_dir, self.tfidf.name, idf_node_name)
        with open(idf_model_file) as idf_model_json:
            idf_model = json.load(idf_model_json)

        idf_shape = {
            'dimensions': [
                {
                    'name': '',
                    'size': 1
                }
            ]
        }

        self.assertEqual(idf_model['attributes']['idf']['double'], [1.0])
        self.assertEqual(idf_model['attributes']['idf']['shape'], idf_shape)
        self.assertEqual(idf_model['attributes']['idf']['type'], 'tensor')
        self.assertEqual(idf_model['op'], 'idf')

    def load_pipeline_model(self):
        with open('{}/{}/root/model.json'.format(self.tmp_dir, self.tfidf.name)) as pipeline_model_json:
            return json.load(pipeline_model_json)



from enum import Enum

from pyspark import keyword_only
from pyspark.ml.param.shared import HasInputCol, HasOutputCol
from pyspark.ml.util import JavaMLReadable
from pyspark.ml.util import JavaMLWritable
from pyspark.ml.util import _jvm
from pyspark.ml.wrapper import JavaTransformer

from mleap.pyspark.py2scala import jvm_scala_object


class UnaryOperation(Enum):
    Sin = 1
    Cos = 2
    Tan = 3
    Log = 4
    Exp = 5
    Abs = 6
    Sqrt = 7
    Logit = 8


class MathUnary(JavaTransformer, HasInputCol, HasOutputCol, JavaMLReadable, JavaMLWritable):

    @keyword_only
    def __init__(self, operation=None, inputCol=None, outputCol=None):
        """
        Computes the mathematical unary `operation` over the input column.

        NOTE: `operation` is not a JavaParam because the underlying 
        MathUnary scala object uses a MathUnaryModel to store the info about
        the unary operation (sin, tan, etc.), not a JavaParam string.

        `operation` has a None default value even though it should *never* be
        None. A None value is necessary upon deserialization to instantiate a
        MathUnary without errors. Afterwards, pyspark sets the _java_obj to
        the deserialized scala object, which encodes the operation.
        """
        super(MathUnary, self).__init__()

        # if operation=None, it means that pyspark is reloading the model
        # from disk and calling this method without args. In such case we don't
        # need to set _java_obj here because pyspark will set it after creation
        #
        # if operation is not None, we can proceed to instantiate the scala classes
        if operation:
            scalaUnaryOperation = jvm_scala_object(
                _jvm().ml.combust.mleap.core.feature,
                f"UnaryOperation${operation.name}$",
            )

            scalaMathUnaryModel = _jvm().ml.combust.mleap.core.feature.MathUnaryModel(scalaUnaryOperation)

            self._java_obj = self._new_java_obj(
                "org.apache.spark.ml.mleap.feature.MathUnary",
                self.uid,
                scalaMathUnaryModel,
            )

        self._setDefault()
        self.setParams(inputCol=inputCol, outputCol=outputCol)

    @keyword_only
    def setParams(self, inputCol=None, outputCol=None):
        """
        Sets params for this MathUnary.
        """
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    def setInputCol(self, value):
        """
        Sets the value of :py:attr:`inputCol`.
        """
        return self._set(inputCol=value)

    def setOutputCol(self, value):
        """
        Sets the value of :py:attr:`outputCol`.
        """
        return self._set(outputCol=value)



from enum import Enum

from pyspark import keyword_only
from pyspark.ml.param.shared import HasInputCol
from pyspark.ml.param.shared import HasOutputCol
from pyspark.ml.param.shared import Param
from pyspark.ml.param.shared import Params
from pyspark.ml.param.shared import TypeConverters
from pyspark.ml.util import JavaMLReadable
from pyspark.ml.util import JavaMLWritable
from pyspark.ml.util import _jvm
from pyspark.ml.wrapper import JavaTransformer

from mleap.pyspark.py2scala import jvm_scala_object
from mleap.pyspark.py2scala import ScalaNone
from mleap.pyspark.py2scala import Some


class BinaryOperation(Enum):
    Add = 1
    Subtract = 2
    Multiply = 3
    Divide = 4
    Remainder = 5
    LogN = 6
    Pow = 7
    Min = 8
    Max = 9


class MathBinary(JavaTransformer, HasOutputCol, JavaMLReadable, JavaMLWritable):

    inputA = Param(
        Params._dummy(),
        "inputA",
        "input for left side of binary operation",
        typeConverter=TypeConverters.toString,
    )

    inputB = Param(
        Params._dummy(),
        "inputB",
        "input for right side of binary operation",
        typeConverter=TypeConverters.toString,
    )

    @keyword_only
    def __init__(
            self,
            operation=None,
            inputA=None,
            inputB=None,
            outputCol=None,
            defaultA=None,
            defaultB=None,
    ):
        """
        Computes the mathematical binary `operation` over
        the input columns A and B.

        :param operation: BinaryOperation to specify the operation type
        :param inputA: column name for the left side of operation (string)
        :param inputB: column name for the right side of operation (string)
        :param outputCol: output column name (string)
        :param defaultA: Default to use instead of inputA. This will only be used
         when inputA is None. For example when defaultA=4,
         operation=BinaryOperation.Multiply and inputB=f1, then all entries of
         col f1 will be multiplied by 4.
        :param defaultB: Default to use instead of inputB. This will only be used
         when inputB is None. For example when defaultB=4,
         operation=BinaryOperation.Multiply and inputA=f1, then all entries of
         col f1 will be multiplied by 4.

        NOTE: `operation`, `defaultA`, `defaultB` is not a JavaParam because
        the underlying MathBinary scala object uses a MathBinaryModel to store
        the info about the binary operation.

        `operation` has a None default value even though it should *never* be
        None. A None value is necessary upon deserialization to instantiate a
        MathBinary without errors. Afterwards, pyspark sets the _java_obj to
        the deserialized scala object, which encodes the operation (as well
        as the default values for A and B).
        """
        super(MathBinary, self).__init__()

        # if operation=None, it means that pyspark is reloading the model
        # from disk and calling this method without args. In such case we don't
        # need to set _java_obj here because pyspark will set it after creation
        #
        # if operation is not None, we can proceed to instantiate the scala classes
        if operation:
            scalaBinaryOperation = jvm_scala_object(
                _jvm().ml.combust.mleap.core.feature,
                f"BinaryOperation${operation.name}$"
            )

            scalaMathBinaryModel = _jvm().ml.combust.mleap.core.feature.MathBinaryModel(
                scalaBinaryOperation,
                Some(defaultA) if defaultA is not None else ScalaNone(),
                Some(defaultB) if defaultB is not None else ScalaNone(),
            )
            self._java_obj = self._new_java_obj(
                "org.apache.spark.ml.mleap.feature.MathBinary",
                self.uid,
                scalaMathBinaryModel,
            )

        self._setDefault()
        self.setParams(inputA=inputA, inputB=inputB, outputCol=outputCol)

    @keyword_only
    def setParams(self, inputA=None, inputB=None, outputCol=None):
        """
        Sets params for this MathBinary.
        """
        # For the correct behavior of MathBinary, params that are None must be unset
        kwargs = {k: v for k, v in self._input_kwargs.items() if v is not None}
        return self._set(**kwargs)

    def setInputA(self, value):
        """
        Sets the value of :py:attr:`inputA`.
        """
        return self._set(inputA=value)

    def setInputB(self, value):
        """
        Sets the value of :py:attr:`inputB`.
        """
        return self._set(inputB=value)

    def setOutputCol(self, value):
        """
        Sets the value of :py:attr:`outputCol`.
        """
        return self._set(outputCol=value)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import shutil
import tempfile
import unittest

import numpy as np
import pandas as pd

from mleap.sklearn.preprocessing.data import FeatureExtractor
from mleap.sklearn.preprocessing.data import SimpleImputer


class TestImputer(unittest.TestCase):
    def setUp(self):
        self.df = pd.DataFrame([
            [0.85281608, 1.50669264],
            [-1.04544152, np.NaN],
            [0.41515407, -0.29941475],
            [np.NaN, -0.96775275],
            [np.NaN, -0.85734022]
        ], columns=['a', 'b'])
        self.feature_extractor = FeatureExtractor(input_scalars=['a'], output_vector='a_extracted')
        self.tmp_dir = tempfile.mkdtemp(prefix="mleap.python.tests")

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_imputer_serialization_fails_with_strategy_set_to_most_frequent(self):
        imputer = SimpleImputer(strategy='most_frequent')
        imputer.mlinit(prior_tf=self.feature_extractor, output_features='a_imputed')

        imputer.fit(self.feature_extractor.transform(self.df))

        with self.assertRaises(NotImplementedError):
            imputer.serialize_to_bundle(self.tmp_dir, imputer.name)

    def test_imputer_serialization_fails_with_strategy_set_to_constant(self):
        imputer = SimpleImputer(strategy='constant')
        imputer.mlinit(prior_tf=self.feature_extractor, output_features='a_imputed')

        imputer.fit(self.feature_extractor.transform(self.df))

        with self.assertRaises(NotImplementedError):
            imputer.serialize_to_bundle(self.tmp_dir, imputer.name)

    def test_imputer_serialization_fails_with_add_indicator_set_to_true(self):
        imputer = SimpleImputer(add_indicator=True)
        imputer.mlinit(prior_tf=self.feature_extractor, output_features='a_imputed')

        imputer.fit(self.feature_extractor.transform(self.df))

        with self.assertRaises(NotImplementedError):
            imputer.serialize_to_bundle(self.tmp_dir, imputer.name)

    def test_imputer_serialization_fails_when_fit_on_multiple_features(self):
        imputer = SimpleImputer()
        self.feature_extractor = FeatureExtractor(input_scalars=['a', 'b'], output_vector='ab_extracted')
        imputer.mlinit(prior_tf=self.feature_extractor, output_features='ab_imputed')

        imputer.fit(self.feature_extractor.transform(self.df))

        with self.assertRaises(NotImplementedError):
            imputer.serialize_to_bundle(self.tmp_dir, imputer.name)

    def test_imputer_serialization_succeeds_with_strategy_set_to_mean(self):
        imputer = SimpleImputer(strategy='mean')
        imputer.mlinit(prior_tf=self.feature_extractor, output_features='a_imputed')

        imputer.fit(self.feature_extractor.transform(self.df))
        imputer.serialize_to_bundle(self.tmp_dir, imputer.name)

        expected_model = {
            "op": "imputer",
            "attributes": {
                "surrogate_value": {
                    "double": self.df.a.mean()
                },
                "strategy": {
                    "string": "mean"
                }
            }
        }

        with open("{}/{}.node/model.json".format(self.tmp_dir, imputer.name)) as json_data:
            actual_model = json.load(json_data)

        self.assertEqual(expected_model, actual_model)

        with open("{}/{}.node/node.json".format(self.tmp_dir, imputer.name)) as json_data:
            node = json.load(json_data)

        self.assertEqual(imputer.name, node['name'])
        self.assertEqual("a_extracted", node['shape']['inputs'][0]['name'])
        self.assertEqual("a_imputed", node['shape']['outputs'][0]['name'])

    def test_imputer_serialization_succeeds_with_strategy_set_to_median(self):
        imputer = SimpleImputer(strategy='median')
        imputer.mlinit(prior_tf=self.feature_extractor, output_features='a_imputed')

        imputer.fit(self.feature_extractor.transform(self.df))
        imputer.serialize_to_bundle(self.tmp_dir, imputer.name)

        expected_model = {
            "op": "imputer",
            "attributes": {
                "surrogate_value": {
                    "double": self.df.a.median()
                },
                "strategy": {
                    "string": "median"
                }
            }
        }

        with open("{}/{}.node/model.json".format(self.tmp_dir, imputer.name)) as json_data:
            actual_model = json.load(json_data)

        self.assertEqual(expected_model, actual_model)

        with open("{}/{}.node/node.json".format(self.tmp_dir, imputer.name)) as json_data:
            node = json.load(json_data)

        self.assertEqual(imputer.name, node['name'])
        self.assertEqual("a_extracted", node['shape']['inputs'][0]['name'])
        self.assertEqual("a_imputed", node['shape']['outputs'][0]['name'])

    def test_imputer_serialization_succeeds_with_missing_values_set_to_zero(self):
        df2 = self.df.fillna(0)

        imputer = SimpleImputer(strategy='mean', missing_values=0.0)
        imputer.mlinit(prior_tf=self.feature_extractor, output_features='a_imputed')

        imputer.fit(self.feature_extractor.transform(df2))
        imputer.serialize_to_bundle(self.tmp_dir, imputer.name)

        expected_model = {
            "op": "imputer",
            "attributes": {
                "surrogate_value": {
                    "double": self.df.a.mean(),
                },
                "strategy": {
                    "string": "mean",
                },
                "missing_value": {
                    "double": 0.0,
                }
            }
        }

        with open("{}/{}.node/model.json".format(self.tmp_dir, imputer.name)) as json_data:
            actual_model = json.load(json_data)

        self.assertEqual(expected_model, actual_model)

        with open("{}/{}.node/node.json".format(self.tmp_dir, imputer.name)) as json_data:
            node = json.load(json_data)

        self.assertEqual(imputer.name, node['name'])
        self.assertEqual("a_extracted", node['shape']['inputs'][0]['name'])
        self.assertEqual("a_imputed", node['shape']['outputs'][0]['name'])



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from sklearn.linear_model import LinearRegression
from mleap.bundle.serialize import MLeapSerializer, MLeapDeserializer
import uuid
import os
import numpy as np


def serialize_to_bundle(self, path, model_name):
    serializer = SimpleSerializer()
    return serializer.serialize_to_bundle(self, path, model_name)


def deserialize_from_bundle(self, path, node_name):
    serializer = SimpleSerializer()
    return serializer.deserialize_from_bundle(self, path, node_name)


def mleap_init(self, input_features, prediction_column):
    self.input_features = input_features
    self.prediction_column = prediction_column
    self.name = "{}_{}".format(self.op, uuid.uuid1())


setattr(LinearRegression, 'op', 'linear_regression')
setattr(LinearRegression, 'mlinit', mleap_init)
setattr(LinearRegression, 'serialize_to_bundle', serialize_to_bundle)
setattr(LinearRegression, 'deserialize_from_bundle', deserialize_from_bundle)
setattr(LinearRegression, 'serializable', True)


class SimpleSerializer(MLeapSerializer, MLeapDeserializer):
    def __init__(self):
        super(SimpleSerializer, self).__init__()

    @staticmethod
    def set_prediction_column(transformer, prediction_column):
        transformer.prediction_column = prediction_column

    @staticmethod
    def set_input_features(transformer, input_features):
        transformer.input_features = input_features

    def serialize_to_bundle(self, transformer, path, model_name):

        # compile tuples of model attributes to serialize
        attributes = list()
        attributes.append(('intercept', transformer.intercept_.tolist()[0]))
        attributes.append(('coefficients', transformer.coef_.tolist()[0]))

        # define node inputs and outputs
        inputs = [{
                  "name": transformer.input_features,
                  "port": "features"
                  }]

        outputs = [{
                  "name": transformer.prediction_column,
                  "port": "prediction"
                   }]

        self.serialize(transformer, path, model_name, attributes, inputs, outputs)

    def deserialize_from_bundle(self, transformer, node_path, node_name):

        attributes_map = {
            'coefficients': 'coef_',
            'intercept': 'intercept_'
        }

        # Set serialized attributes
        full_node_path = os.path.join(node_path, node_name)
        transformer = self.deserialize_single_input_output(transformer, full_node_path, attributes_map)

        # Set Additional Attributes
        if 'intercept_' in transformer.__dict__:
            transformer.fit_intercept = True
        else:
            transformer.fit_intercept = False

        transformer.coef_ = np.array([transformer.coef_])

        return transformer




#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
import pandas as pd
import numpy as np
import os
import shutil
import json
import uuid
import tempfile

from sklearn.linear_model import LinearRegression, LogisticRegression, LogisticRegressionCV
from sklearn.preprocessing import Binarizer
from mleap.sklearn.base import LinearRegression
from mleap.sklearn.logistic import LogisticRegression, LogisticRegressionCV
from mleap.sklearn.preprocessing.data import FeatureExtractor, Binarizer


def to_standard_normal_quartile(rand_num):
    """Retrieve the quartile of a data point sampled from the standard normal distribution

    Useful for assigning multi-class labels to random data during tests
    Such tests should probably use sklearn.preprocessing.KBinsDiscretizer instead
    But they can't since scikit-learn is pinned < 0.20.0
    https://github.com/combust/mleap/pull/431

    """
    if rand_num < -0.67448:
        return 0
    elif rand_num < 0:
        return 1
    elif rand_num < 0.67448:
        return 2
    else:
        return 3


class TransformerTests(unittest.TestCase):
    def setUp(self):
        self.df = pd.DataFrame(np.random.randn(100, 5), columns=['a', 'b', 'c', 'd', 'e'])
        self.tmp_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_linear_regression_serializer(self):

        linear_regression = LinearRegression(fit_intercept=True, normalize=False)
        linear_regression.mlinit(input_features='a',
                                 prediction_column='e')

        linear_regression.fit(self.df[['a']], self.df[['e']])

        linear_regression.serialize_to_bundle(self.tmp_dir, linear_regression.name)


        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, linear_regression.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual(model['op'], 'linear_regression')
        self.assertTrue(model['attributes']['intercept']['double'] is not None)

    def test_linear_regression_deserializer(self):

        linear_regression = LinearRegression(fit_intercept=True, normalize=False)
        linear_regression.mlinit(input_features='a',
                                 prediction_column='e')

        linear_regression.fit(self.df[['a']], self.df[['e']])

        linear_regression.serialize_to_bundle(self.tmp_dir, linear_regression.name)

        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, linear_regression.name)) as json_data:
            model = json.load(json_data)

        # Now deserialize it back
        node_name = "{}.node".format(linear_regression.name)
        linear_regression_tf = LinearRegression()
        linear_regression_tf = linear_regression_tf.deserialize_from_bundle(self.tmp_dir, node_name)

        res_a = linear_regression.predict(self.df[['a']])
        res_b = linear_regression_tf.predict(self.df[['a']])

        self.assertEqual(res_a[0], res_b[0])
        self.assertEqual(res_a[1], res_b[1])
        self.assertEqual(res_a[2], res_b[2])

    def test_logistic_regression_serializer(self):

        logistic_regression = LogisticRegression(fit_intercept=True)
        logistic_regression.mlinit(input_features='a',
                                 prediction_column='e_binary')

        extract_features = ['e']
        feature_extractor = FeatureExtractor(input_scalars=['e'],
                                         output_vector='extracted_e_output',
                                         output_vector_items=["{}_out".format(x) for x in extract_features])

        binarizer = Binarizer(threshold=0.0)
        binarizer.mlinit(prior_tf=feature_extractor,
                         output_features='e_binary')

        Xres = binarizer.fit_transform(self.df[['a']])

        logistic_regression.fit(self.df[['a']], Xres)

        logistic_regression.serialize_to_bundle(self.tmp_dir, logistic_regression.name)


        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, logistic_regression.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual(model['op'], 'logistic_regression')
        self.assertTrue(model['attributes']['intercept']['double'] is not None)

    def test_logistic_regression_deserializer(self):

        logistic_regression = LogisticRegression(fit_intercept=True)
        logistic_regression.mlinit(input_features='a',
                                   prediction_column='e_binary')

        extract_features = ['e']
        feature_extractor = FeatureExtractor(input_scalars=['e'],
                                         output_vector='extracted_e_output',
                                         output_vector_items=["{}_out".format(x) for x in extract_features])

        binarizer = Binarizer(threshold=0.0)
        binarizer.mlinit(prior_tf=feature_extractor,
                         output_features='e_binary')

        Xres = binarizer.fit_transform(self.df[['a']])

        logistic_regression.fit(self.df[['a']], Xres)

        logistic_regression.serialize_to_bundle(self.tmp_dir, logistic_regression.name)

        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, logistic_regression.name)) as json_data:
            model = json.load(json_data)

        # Now deserialize it back
        node_name = "{}.node".format(logistic_regression.name)
        logistic_regression_tf = LogisticRegression()
        logistic_regression_tf = logistic_regression_tf.deserialize_from_bundle(self.tmp_dir, node_name)

        res_a = logistic_regression.predict(self.df[['a']])
        res_b = logistic_regression_tf.predict(self.df[['a']])

        self.assertEqual(res_a[0], res_b[0])
        self.assertEqual(res_a[1], res_b[1])
        self.assertEqual(res_a[2], res_b[2])

    def test_multinomial_logistic_regression_serializer(self):

        logistic_regression = LogisticRegression(fit_intercept=True)
        logistic_regression.mlinit(
            input_features='a',
            prediction_column='prediction'
        )

        X = self.df[['a']]
        y = np.array([to_standard_normal_quartile(elem) for elem in X.to_numpy()])

        logistic_regression.fit(X, y)
        logistic_regression.serialize_to_bundle(self.tmp_dir, logistic_regression.name)

        with open("{}/{}.node/model.json".format(self.tmp_dir, logistic_regression.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual(model['op'], 'logistic_regression')
        self.assertEqual(model['attributes']['num_classes']['long'], 4)

        # assert 4x1 coefficient matrix
        self.assertEqual(len(model['attributes']['coefficient_matrix']['double']), 4)
        self.assertEqual(len(model['attributes']['coefficient_matrix']['shape']['dimensions']), 2)
        self.assertEqual(model['attributes']['coefficient_matrix']['shape']['dimensions'][0]['size'], 4)
        self.assertEqual(model['attributes']['coefficient_matrix']['shape']['dimensions'][1]['size'], 1)

        # assert 4x0 intercept vector
        self.assertEqual(len(model['attributes']['intercept_vector']['double']), 4)
        self.assertEqual(len(model['attributes']['intercept_vector']['shape']['dimensions']), 1)
        self.assertEqual(model['attributes']['intercept_vector']['shape']['dimensions'][0]['size'], 4)

    def test_multinomial_logistic_regression_deserializer(self):

        logistic_regression = LogisticRegression(fit_intercept=True)
        logistic_regression.mlinit(
            input_features='a',
            prediction_column='prediction'
        )

        X = self.df[['a']]
        y = np.array([to_standard_normal_quartile(elem) for elem in X.to_numpy()])

        logistic_regression.fit(X, y)
        logistic_regression.serialize_to_bundle(self.tmp_dir, logistic_regression.name)

        node_name = "{}.node".format(logistic_regression.name)
        logistic_regression_tf = LogisticRegression()
        logistic_regression_tf = logistic_regression_tf.deserialize_from_bundle(self.tmp_dir, node_name)

        expected = logistic_regression.predict(self.df[['a']])
        actual = logistic_regression_tf.predict(self.df[['a']])

        np.testing.assert_array_equal(expected, actual)

    def test_logistic_regression_cv_serializer(self):

        logistic_regression = LogisticRegressionCV(fit_intercept=True)
        logistic_regression.mlinit(input_features='a',
                                 prediction_column='e_binary')

        extract_features = ['e']
        feature_extractor = FeatureExtractor(input_scalars=['e'],
                                             output_vector='extracted_e_output',
                                             output_vector_items=["{}_out".format(x) for x in extract_features])

        binarizer = Binarizer(threshold=0.0)
        binarizer.mlinit(prior_tf=feature_extractor,
                         output_features='e_binary')

        Xres = binarizer.fit_transform(self.df[['a']])

        logistic_regression.fit(self.df[['a']], Xres)

        logistic_regression.serialize_to_bundle(self.tmp_dir, logistic_regression.name)


        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, logistic_regression.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual(model['op'], 'logistic_regression')
        self.assertTrue(model['attributes']['intercept']['double'] is not None)

    def test_logistic_regression_cv_deserializer(self):

        logistic_regression = LogisticRegressionCV(fit_intercept=True)
        logistic_regression.mlinit(input_features='a',
                                   prediction_column='e_binary')

        extract_features = ['e']
        feature_extractor = FeatureExtractor(input_scalars=['e'],
                                             output_vector='extracted_e_output',
                                             output_vector_items=["{}_out".format(x) for x in extract_features])

        binarizer = Binarizer(threshold=0.0)
        binarizer.mlinit(prior_tf=feature_extractor,
                         output_features='e_binary')

        Xres = binarizer.fit_transform(self.df[['a']])

        logistic_regression.fit(self.df[['a']], Xres)

        logistic_regression.serialize_to_bundle(self.tmp_dir, logistic_regression.name)

        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, logistic_regression.name)) as json_data:
            model = json.load(json_data)

        # Now deserialize it back
        node_name = "{}.node".format(logistic_regression.name)
        logistic_regression_tf = LogisticRegressionCV()
        logistic_regression_tf = logistic_regression_tf.deserialize_from_bundle(self.tmp_dir, node_name)

        res_a = logistic_regression.predict(self.df[['a']])
        res_b = logistic_regression_tf.predict(self.df[['a']])

        self.assertEqual(res_a[0], res_b[0])
        self.assertEqual(res_a[1], res_b[1])
        self.assertEqual(res_a[2], res_b[2])



import re
from pandas.testing import assert_frame_equal


def _print_dfs(actual, expected):
    print('actual.show(100):')
    actual.show(100, False)
    print('expected.show(100):')
    expected.show(100, False)


def assert_df(actual, expected, sort=True):
    """
    For more elaborate DFs comparison, e.g. ignoring sorting or less precise.
    Default precision for float comparison is 5 digits.
    """
    __tracebackhide__ = True

    if sort:
        actual = actual.orderBy(actual.columns)
        expected = expected.orderBy(actual.columns)

    try:
        assert_frame_equal(actual.toPandas(), expected.toPandas())
    except AssertionError as e:
        _print_dfs(actual, expected)
        # assert_frame_equal doesn't print the column name, only the index
        #   -> get the index from message with regex & resolve column name
        matcher = re.search(r'iloc\[:, (\d+)\]', e.args[0])
        if matcher:
            iloc = matcher.group(1)
            if iloc is not None:
                raise AssertionError("failed assert on column '{}': {}".format(actual.columns[int(iloc)], e))
        # couldn't extract column name. unexpected.
        raise e



import json
import numpy as np
import pandas as pd
import shutil
import tempfile
import unittest

from mleap.sklearn.extensions.data import Imputer

class ExtensionsTests(unittest.TestCase):
    def setUp(self):
        self.df = pd.DataFrame(np.random.randn(10, 5), columns=['a', 'b', 'c', 'd', 'e'])
        self.tmp_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def imputer_ext_test(self):

        def _set_nulls(df):
            row = df['index']
            if row in [2,5]:
                return np.NaN
            return df.a

        imputer = Imputer(strategy='mean', input_features='a', output_features='a_imputed')

        df2 = self.df
        df2.reset_index(inplace=True)
        df2['a'] = df2.apply(_set_nulls, axis=1)

        imputer.fit(df2)

        self.assertAlmostEqual(imputer.statistics_[0], df2.a.mean(), places = 7)

        imputer.serialize_to_bundle(self.tmp_dir, imputer.name)

        expected_model = {
            "op": "imputer",
            "attributes": {
                "surrogate_value": {
                    "double": df2.a.mean()
                },
                "strategy": {
                    "string": "mean"
                }
            }
        }

        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, imputer.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual(expected_model['attributes']['strategy']['string'], model['attributes']['strategy']['string'])
        self.assertAlmostEqual(expected_model['attributes']['surrogate_value']['double'], model['attributes']['surrogate_value']['double'], places = 7)

        # Test node.json
        with open("{}/{}.node/node.json".format(self.tmp_dir, imputer.name)) as json_data:
            node = json.load(json_data)

        self.assertEqual(imputer.name, node['name'])
        self.assertEqual(imputer.input_features, node['shape']['inputs'][0]['name'])
        self.assertEqual(imputer.output_features, node['shape']['outputs'][0]['name'])



import os
import shutil
import six
import json
import numpy as np

_type_map = {
    int: 'long',
    float: 'double',
    str: 'string',
    np.float64: 'double'
}


class Vector(object):
    def __init__(self, values):
        self.values = values


class MLeapSerializer(object):
    """
    Base class to serialize transformers and estimators to a bundle.ml file. Main components that get serialized are:
        - Model: Contains the data needed for the transformer. For example, if the transformer is a linear regression,
                then we serialize the coefficients and the intercept of the model.
        - Node: Contains the definition of the input and output data.
    """
    def __init__(self):
        pass

    def get_mleap_model(self, transformer, attributes_to_serialize=None):
        """
        Generates the model.json given a list of attributes, which are a tuple comprised of:
            - name
            - value
        Type is figured out automatically, but we should consider specifying it explicitly.
        Note: this only supports doubles and tensors that are vectors/lists of doubles.
        :param transformer:
        :param attributes_to_serialize: Tuple of (name, value)
        :return:
        """
        js = {
            'op': transformer.op
        }

        # If the transformer doesn't have any attributes, return just the op name
        if attributes_to_serialize is None:
            return js

        attributes = {}

        for name, value in attributes_to_serialize:
            if isinstance(value, float):
                attributes[name] = {
                    "double": value
                }

            elif isinstance(value, bool) and value in [True, False]:
                attributes[name] = {
                    "boolean": value
                }

            elif isinstance(value, int):
                attributes[name] = {
                    "long": value
                }
            elif isinstance(value, Vector):
                attributes[name] = {
                    "type": "list",
                    "double": value.values
                }
            elif isinstance(value, list) and (isinstance(value[0], np.float64) or isinstance(value[0], np.float32) or isinstance(value[0], float)):
                base = type(value[0])
                attributes[name] = {
                    _type_map[base]: value,
                    "shape": {
                        "dimensions": [{
                            "size": len(value),
                            "name": ""
                        }]
                    },
                    "type": "tensor"
                }
            elif isinstance(value, list) and isinstance(value[0], str):
                attributes[name] = {
                    "type": "list",
                    "string": value
                }

            elif isinstance(value, np.ndarray):
                attributes[name] = {
                    "double": value.flatten().tolist(),
                    "shape": {
                        "dimensions": [{
                            "size": dim,
                            "name": ""
                        } for dim in value.shape]
                    },
                    "type": "tensor"
                }

            elif isinstance(value, str):
                attributes[name] = {
                    'string': value
                }

            elif isinstance(value, dict):
                shapes = list()
                for shape in value['data_shape']:
                    if shape['shape'] == 'scalar':
                        shapes.append(({"base": "scalar",
                                        "isNullable": False}))
                    elif shape['shape'] == 'tensor':
                        shapes.append(({
                            "base": "tensor",
                            "isNullable": False,
                            "tensorShape": {
                                "dimensions": [{
                                    "size": shape['tensor_shape']['dimensions'][0]['size'],
                                    "name": ""
                                }]
                            }
                        }))
                attributes[name] = {
                    'type': 'list',
                    'data_shape': shapes
                }

        js['attributes'] = attributes

        return js

    def get_mleap_node(self, transformer, inputs, outputs):
        js = {
            "name": transformer.name,
            "shape": {
                "inputs": inputs,
                "outputs": outputs
            }
        }
        return js

    def serialize(self, transformer, path, model_name, attributes, inputs, outputs, node=True, model=True):
        # If bundle path already exists, delete it and create a clean directory
        if node:
            if os.path.exists("{}/{}.node".format(path, model_name)):
                shutil.rmtree("{}/{}.node".format(path, model_name))

            model_dir = "{}/{}.node".format(path, model_name)
        else:
            if os.path.exists("{}/{}".format(path, model_name)):
                shutil.rmtree("{}/{}".format(path, model_name))

            model_dir = "{}/{}".format(path, model_name)

        os.mkdir(model_dir)

        if model:
            # Write bundle file
            with open("{}/{}".format(model_dir, 'model.json'), 'w') as outfile:
                json.dump(self.get_mleap_model(transformer, attributes), outfile, indent=3)

        if node:
            # Write node file
            with open("{}/{}".format(model_dir, 'node.json'), 'w') as outfile:
                json.dump(self.get_mleap_node(transformer, inputs, outputs), outfile, indent=3)


class MLeapDeserializer(object):

    def deserialize_from_bundle(self, transformer, node_path, node_name):
        """
        :type node_path: str
        :type node_name: str
        :type transformer: StandardScaler
        :param transformer:
        :param node_path:
        :param node_name:
        :return:
        """
        NotImplementedError()

    @staticmethod
    def _node_features_format(x):
        if isinstance(x, six.string_types):
            return [str(x)]
        return x

    def deserialize_single_input_output(self, transformer, node_path, attributes_map=None):
        """
        :attributes_map: Map of attributes names. For example StandardScaler has `mean_` but is serialized as `mean`
        :param transformer: Scikit or Pandas transformer
        :param node: bundle.ml node json file
        :param model: bundle.ml model json file
        :return: Transformer
        """
        # Load the model file
        with open("{}/model.json".format(node_path)) as json_data:
            model_j = json.load(json_data)

        # Set Transformer Attributes
        attributes = model_j['attributes']
        for attribute in attributes.keys():
            value_key = [key for key in attributes[attribute].keys()
                         if key in ['string', 'boolean', 'long', 'double', 'data_shape']][0]
            if attributes_map is not None and attribute in attributes_map.keys():
                setattr(transformer, attributes_map[attribute], attributes[attribute][value_key])
            else:
                setattr(transformer, attribute, attributes[attribute][value_key])

        transformer.op = model_j['op']

        # Load the node file
        with open("{}/node.json".format(node_path)) as json_data:
            node_j = json.load(json_data)

        transformer.name = node_j['name']
        transformer.input_features = self._node_features_format(node_j['shape']['inputs'][0]['name'])
        transformer.output_features = self._node_features_format(node_j['shape']['outputs'][0]['name'])

        return transformer



class NodeLookup(object):
    """Converts integer node ID's to human readable labels."""

    def __init__(self,
                 label_lookup_path=None,
                 uid_lookup_path=None):
        if not label_lookup_path:
            label_lookup_path = os.path.join(
                    FLAGS.model_dir, 'imagenet_2012_challenge_label_map_proto.pbtxt')
        if not uid_lookup_path:
            uid_lookup_path = os.path.join(
                    FLAGS.model_dir, 'imagenet_synset_to_human_label_map.txt')
        self.node_lookup = self.load(label_lookup_path, uid_lookup_path)

    def load(self, label_lookup_path, uid_lookup_path):
        """Loads a human readable English name for each softmax node.
        Args:
          label_lookup_path: string UID to integer node ID.
          uid_lookup_path: string UID to human-readable string.
        Returns:
          dict from integer node ID to human-readable string.
        """
        if not tf.gfile.Exists(uid_lookup_path):
            tf.logging.fatal('File does not exist %s', uid_lookup_path)
        if not tf.gfile.Exists(label_lookup_path):
            tf.logging.fatal('File does not exist %s', label_lookup_path)

        # Loads mapping from string UID to human-readable string
        proto_as_ascii_lines = tf.gfile.GFile(uid_lookup_path).readlines()
        uid_to_human = {}
        p = re.compile(r'[n\d]*[ \S,]*')
        for line in proto_as_ascii_lines:
            parsed_items = p.findall(line)
            uid = parsed_items[0]
            human_string = parsed_items[2]
            uid_to_human[uid] = human_string

        # Loads mapping from string UID to integer node ID.
        node_id_to_uid = {}
        proto_as_ascii = tf.gfile.GFile(label_lookup_path).readlines()
        for line in proto_as_ascii:
            if line.startswith('  target_class:'):
                target_class = int(line.split(': ')[1])
            if line.startswith('  target_class_string:'):
                target_class_string = line.split(': ')[1]
                node_id_to_uid[target_class] = target_class_string[1:-2]

        # Loads the final mapping of integer node ID to human-readable string
        node_id_to_name = {}
        for key, val in node_id_to_uid.items():
            if val not in uid_to_human:
                tf.logging.fatal('Failed to locate: %s', val)
            name = uid_to_human[val]
            node_id_to_name[key] = name

        return node_id_to_name

    def id_to_string(self, node_id):
        if node_id not in self.node_lookup:
            return ''
        return self.node_lookup[node_id]



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import shutil
import tempfile
import unittest

import numpy as np
import pandas as pd
from mleap.sklearn.preprocessing.data import FeatureExtractor, MathUnary, MathBinary, StringMap
from mleap.sklearn.preprocessing.data import StandardScaler, MinMaxScaler, LabelEncoder, Binarizer, PolynomialFeatures
from pandas.testing import assert_frame_equal


class TransformerTests(unittest.TestCase):
    def setUp(self):
        self.df = pd.DataFrame(np.random.randn(10, 5), columns=['a', 'b', 'c', 'd', 'e'])
        self.tmp_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_standard_scaler_serializer(self):

        standard_scaler = StandardScaler(with_mean=True,
                                         with_std=True
                                         )

        extract_features = ['a']
        feature_extractor = FeatureExtractor(input_scalars=['a'],
                                             output_vector='extracted_a_output',
                                             output_vector_items=["{}_out".format(x) for x in extract_features])

        standard_scaler.mlinit(prior_tf=feature_extractor,
                               output_features='a_scaled')

        standard_scaler.fit(self.df[['a']])

        standard_scaler.serialize_to_bundle(self.tmp_dir, standard_scaler.name)

        expected_mean = self.df.a.mean()
        expected_std = np.sqrt(np.var(self.df.a))

        expected_model = {
            "op": "standard_scaler",
            "attributes": {
                "mean": {
                    "double": [expected_mean],
                    "shape": {
                        "dimensions": [{
                            "size": 1,
                            "name": ""
                        }]
                    },
                    "type": "tensor"
                },
                "std": {
                    "double": [expected_std],
                    "shape": {
                        "dimensions": [{
                            "size": 1,
                            "name": ""
                        }]
                    },
                    "type": "tensor"
                }
            }
        }

        self.assertAlmostEqual(expected_mean, standard_scaler.mean_.tolist()[0], places = 7)
        self.assertAlmostEqual(expected_std, np.sqrt(standard_scaler.var_.tolist()[0]), places = 7)

        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, standard_scaler.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual(standard_scaler.op, expected_model['op'])
        self.assertEqual(expected_model['attributes']['mean']['shape']['dimensions'][0]['size'], model['attributes']['mean']['shape']['dimensions'][0]['size'])
        self.assertEqual(expected_model['attributes']['std']['shape']['dimensions'][0]['size'], model['attributes']['std']['shape']['dimensions'][0]['size'])
        self.assertAlmostEqual(expected_model['attributes']['mean']['double'][0], model['attributes']['mean']['double'][0], places = 7)
        self.assertAlmostEqual(expected_model['attributes']['std']['double'][0], model['attributes']['std']['double'][0], places = 7)

        # Test node.json
        with open("{}/{}.node/node.json".format(self.tmp_dir, standard_scaler.name)) as json_data:
            node = json.load(json_data)

        self.assertEqual(standard_scaler.name, node['name'])
        self.assertEqual(standard_scaler.input_features, node['shape']['inputs'][0]['name'])
        self.assertEqual(standard_scaler.output_features, node['shape']['outputs'][0]['name'])

    def test_standard_scaler_deserializer(self):

        extract_features = ['a']
        feature_extractor = FeatureExtractor(input_scalars=['a'],
                                         output_vector='extracted_a_output',
                                         output_vector_items=["{}_out".format(x) for x in extract_features])

        # Serialize a standard scaler to a bundle
        standard_scaler = StandardScaler(with_mean=True,
                                         with_std=True
                                         )

        standard_scaler.mlinit(prior_tf=feature_extractor,
                               output_features='a_scaled')

        standard_scaler.fit(self.df[['a']])

        standard_scaler.serialize_to_bundle(self.tmp_dir, standard_scaler.name)

        # Now deserialize it back

        node_name = "{}.node".format(standard_scaler.name)

        standard_scaler_tf = StandardScaler()

        standard_scaler_tf = standard_scaler_tf.deserialize_from_bundle(self.tmp_dir, node_name)

        # Transform some sample data
        res_a = standard_scaler.transform(self.df[['a']])
        res_b = standard_scaler_tf.transform(self.df[['a']])

        self.assertEqual(res_a[0], res_b[0])
        self.assertEqual(standard_scaler.name, standard_scaler_tf.name)
        self.assertEqual(standard_scaler.op, standard_scaler_tf.op)
        self.assertEqual(standard_scaler.mean_, standard_scaler_tf.mean_)
        self.assertEqual(standard_scaler.scale_, standard_scaler_tf.scale_)

    def test_standard_scaler_multi_deserializer(self):

        extract_features = ['a', 'b']
        feature_extractor = FeatureExtractor(input_scalars=['a', 'b'],
                                             output_vector='extracted_multi_outputs',
                                             output_vector_items=["{}_out".format(x) for x in extract_features])

        # Serialize a standard scaler to a bundle
        standard_scaler = StandardScaler(with_mean=True,
                                         with_std=True
                                         )

        standard_scaler.mlinit(prior_tf=feature_extractor,
                               output_features=['a_scaled', 'b_scaled'])

        standard_scaler.fit(self.df[['a', 'b']])

        standard_scaler.serialize_to_bundle(self.tmp_dir, standard_scaler.name)

        # Now deserialize it back

        node_name = "{}.node".format(standard_scaler.name)

        standard_scaler_tf = StandardScaler()

        standard_scaler_tf = standard_scaler_tf.deserialize_from_bundle(self.tmp_dir, node_name)

        # Transform some sample data
        res_a = standard_scaler.transform(self.df[['a', 'b']])
        res_b = standard_scaler_tf.transform(self.df[['a', 'b']])

        self.assertEqual(res_a[0][0], res_b[0][0])
        self.assertEqual(res_a[0][1], res_b[0][1])
        self.assertEqual(standard_scaler.name, standard_scaler_tf.name)
        self.assertEqual(standard_scaler.op, standard_scaler_tf.op)
        self.assertEqual(standard_scaler.mean_[0], standard_scaler_tf.mean_[0])
        self.assertEqual(standard_scaler.mean_[1], standard_scaler_tf.mean_[1])
        self.assertEqual(standard_scaler.scale_[0], standard_scaler_tf.scale_[0])
        self.assertEqual(standard_scaler.scale_[1], standard_scaler_tf.scale_[1])

    def test_min_max_scaler_serializer(self):

        extract_features = ['a']
        feature_extractor = FeatureExtractor(input_scalars=['a'],
                                         output_vector='extracted_a_output',
                                         output_vector_items=["{}_out".format(x) for x in extract_features])

        scaler = MinMaxScaler()
        scaler.mlinit(prior_tf = feature_extractor,
                      output_features='a_scaled')

        scaler.fit(self.df[['a']])

        scaler.serialize_to_bundle(self.tmp_dir, scaler.name)

        expected_min = self.df.a.min()
        expected_max = self.df.a.max()

        expected_model = {
           "op": "min_max_scaler",
            "attributes": {
                "min": {
                    "double": [expected_min],
                    "shape": {
                        "dimensions": [{
                            "size": 1,
                            "name": ""
                        }]
                    },
                    "type": "tensor"
                },
                "max": {
                    "double": [expected_max],
                    "shape": {
                        "dimensions": [{
                            "size": 1,
                            "name": ""
                        }]
                    },
                    "type": "tensor"
                }
            }
        }

        self.assertEqual(expected_min, scaler.data_min_.tolist()[0])
        self.assertEqual(expected_max, scaler.data_max_.tolist()[0])

        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, scaler.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual(scaler.op, expected_model['op'])
        self.assertEqual(expected_model['attributes']['min']['shape']['dimensions'][0]['size'], model['attributes']['min']['shape']['dimensions'][0]['size'])
        self.assertEqual(expected_model['attributes']['max']['shape']['dimensions'][0]['size'], model['attributes']['max']['shape']['dimensions'][0]['size'])
        self.assertEqual(expected_model['attributes']['min']['double'][0], model['attributes']['min']['double'][0])
        self.assertEqual(expected_model['attributes']['max']['double'][0], model['attributes']['max']['double'][0])

        # Test node.json
        with open("{}/{}.node/node.json".format(self.tmp_dir, scaler.name)) as json_data:
            node = json.load(json_data)

        self.assertEqual(scaler.name, node['name'])
        self.assertEqual(scaler.input_features, node['shape']['inputs'][0]['name'])
        self.assertEqual(scaler.output_features, node['shape']['outputs'][0]['name'])

    def test_min_max_scaler_deserializer(self):

        extract_features = ['a']
        feature_extractor = FeatureExtractor(input_scalars=['a'],
                                             output_vector='extracted_a_output',
                                             output_vector_items=["{}_out".format(x) for x in extract_features])

        scaler = MinMaxScaler()
        scaler.mlinit(prior_tf=feature_extractor,
                      output_features='a_scaled')

        scaler.fit(self.df[['a']])

        scaler.serialize_to_bundle(self.tmp_dir, scaler.name)

        # Deserialize the MinMaxScaler
        node_name = "{}.node".format(scaler.name)
        min_max_scaler_tf = MinMaxScaler()
        min_max_scaler_tf.deserialize_from_bundle(self.tmp_dir, node_name)

        # Transform some sample data
        res_a = scaler.transform(self.df[['a']])
        res_b = min_max_scaler_tf.transform(self.df[['a']])

        self.assertEqual(res_a[0], res_b[0])

        self.assertEqual(scaler.name, min_max_scaler_tf.name)
        self.assertEqual(scaler.op, min_max_scaler_tf.op)

    def test_min_max_scaler_multi_deserializer(self):

        extract_features = ['a', 'b']
        feature_extractor = FeatureExtractor(input_scalars=extract_features,
                                             output_vector='extracted_multi_outputs',
                                             output_vector_items=["{}_out".format(x) for x in extract_features])

        scaler = MinMaxScaler()
        scaler.mlinit(prior_tf=feature_extractor,
                      output_features=['a_scaled', 'b_scaled'])

        scaler.fit(self.df[extract_features])

        scaler.serialize_to_bundle(self.tmp_dir, scaler.name)

        # Deserialize the MinMaxScaler
        node_name = "{}.node".format(scaler.name)
        min_max_scaler_tf = MinMaxScaler()
        min_max_scaler_tf.deserialize_from_bundle(self.tmp_dir, node_name)

        # Transform some sample data
        res_a = scaler.transform(self.df[extract_features])
        res_b = min_max_scaler_tf.transform(self.df[extract_features])

        self.assertEqual(res_a[0][0], res_b[0][0])
        self.assertEqual(res_a[0][1], res_b[0][1])

        self.assertEqual(scaler.name, min_max_scaler_tf.name)
        self.assertEqual(scaler.op, min_max_scaler_tf.op)

    def label_encoder_test(self):

        labels = ['a', 'b', 'c']

        le = LabelEncoder(input_features=['label_feature'],
                  output_features='label_feature_le_encoded')

        le.fit(labels)

        self.assertEqual(labels, le.classes_.tolist())

        le.serialize_to_bundle(self.tmp_dir, le.name)

        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, le.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual(le.op, model['op'])
        self.assertTrue('nullable_input' in model['attributes'])
        self.assertTrue('labels' in model['attributes'])

        # Test node.json
        with open("{}/{}.node/node.json".format(self.tmp_dir, le.name)) as json_data:
            node = json.load(json_data)

        self.assertEqual(le.name, node['name'])
        self.assertEqual(le.input_features[0], node['shape']['inputs'][0]['name'])
        self.assertEqual(le.output_features, node['shape']['outputs'][0]['name'])

    def label_encoder_deserializer_test(self):

        labels = ['a', 'b', 'c']
        le = LabelEncoder(input_features=['label_feature'],
                          output_features='label_feature_le_encoded')

        le.fit(labels)

        self.assertEqual(labels, le.classes_.tolist())

        le.serialize_to_bundle(self.tmp_dir, le.name)

        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, le.name)) as json_data:
            model = json.load(json_data)

        # Deserialize the LabelEncoder
        node_name = "{}.node".format(le.name)
        label_encoder_tf = LabelEncoder()
        label_encoder_tf.deserialize_from_bundle(self.tmp_dir, node_name)

        # Transform some sample data
        res_a = le.transform(labels)
        res_b = label_encoder_tf.transform(labels)
        print("le.output_features: {}".format(le.output_features))
        print("label_encoder_tf.output_features: {}".format(label_encoder_tf.output_features))
        self.assertEqual(res_a[0], res_b[0])
        self.assertEqual(res_a[1], res_b[1])
        self.assertEqual(res_a[2], res_b[2])
        self.assertEqual(le.input_features, label_encoder_tf.input_features)
        self.assertEqual(le.output_features, label_encoder_tf.output_features[0])

    def feature_extractor_test(self):

        extract_features = ['a', 'd']

        feature_extractor = FeatureExtractor(input_scalars=extract_features,
                                             output_vector='extract_features_output',
                                             output_vector_items=["{}_out".format(x) for x in extract_features])

        res = feature_extractor.fit_transform(self.df)

        self.assertEqual(len(res.columns), 2)

        feature_extractor.serialize_to_bundle(self.tmp_dir, feature_extractor.name)

        # Test node.json
        with open("{}/{}.node/node.json".format(self.tmp_dir, feature_extractor.name)) as json_data:
            node = json.load(json_data)

        self.assertEqual(feature_extractor.name, node['name'])
        self.assertEqual(feature_extractor.input_features[0], node['shape']['inputs'][0]['name'])
        self.assertEqual(feature_extractor.input_features[1], node['shape']['inputs'][1]['name'])
        self.assertEqual(feature_extractor.output_vector, node['shape']['outputs'][0]['name'])

        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, feature_extractor.name)) as json_data:
            model = json.load(json_data)

        expected_model = {
            "op": "vector_assembler",
            "attributes": {
                "input_shapes": {
                    "data_shape": [
                        {
                        "base": "scalar",
                        "isNullable": False
                        },
                        {
                        "base": "scalar",
                        "isNullable": False
                        }],
                    "type": "list"
                }
            }
        }

        self.assertEqual(expected_model['op'], model['op'])
        self.assertEqual(expected_model['attributes']['input_shapes']['data_shape'][0]['base'],
                         model['attributes']['input_shapes']['data_shape'][0]['base'])
        self.assertEqual(expected_model['attributes']['input_shapes']['data_shape'][0]['isNullable'],
                         model['attributes']['input_shapes']['data_shape'][0]['isNullable'])
        self.assertEqual(expected_model['attributes']['input_shapes']['data_shape'][1]['base'],
                         model['attributes']['input_shapes']['data_shape'][1]['base'])
        self.assertEqual(expected_model['attributes']['input_shapes']['data_shape'][1]['isNullable'],
                     model['attributes']['input_shapes']['data_shape'][1]['isNullable'])

    def binarizer_test(self):

        extract_features = ['a']
        feature_extractor = FeatureExtractor(input_scalars=['a'],
                                         output_vector='extracted_a_output',
                                         output_vector_items=["{}_out".format(x) for x in extract_features])

        binarizer = Binarizer(threshold=0)
        binarizer.mlinit(prior_tf=feature_extractor,
                         output_features='a_binary')

        Xres = binarizer.fit_transform(self.df[['a']])

        # Test that the binarizer functions as expected
        self.assertEqual(float(len(self.df[self.df.a >= 0]))/10.0, Xres.mean())

        binarizer.serialize_to_bundle(self.tmp_dir, binarizer.name)

        expected_model = {
          "op": "sklearn_binarizer",
          "attributes": {
            "threshold": {
              "double": 0.0
            }
          }
        }

        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, binarizer.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual(expected_model['attributes']['threshold']['double'],
                         model['attributes']['threshold']['double'])
        self.assertEqual(expected_model['op'], model['op'])

        # Test node.json
        with open("{}/{}.node/node.json".format(self.tmp_dir, binarizer.name)) as json_data:
            node = json.load(json_data)

        self.assertEqual(binarizer.name, node['name'])
        self.assertEqual(binarizer.input_features, node['shape']['inputs'][0]['name'])
        self.assertEqual(binarizer.output_features, node['shape']['outputs'][0]['name'])

    def binarizer_deserializer_test(self):

        extract_features = ['a']
        feature_extractor = FeatureExtractor(input_scalars=['a'],
                                         output_vector='extracted_a_output',
                                         output_vector_items=["{}_out".format(x) for x in extract_features])

        binarizer = Binarizer(threshold=0.0)
        binarizer.mlinit(prior_tf=feature_extractor,
                         output_features='a_binary')

        Xres = binarizer.fit_transform(self.df[['a']])

        # Test that the binarizer functions as expected
        self.assertEqual(float(len(self.df[self.df.a >= 0]))/10.0, Xres.mean())

        binarizer.serialize_to_bundle(self.tmp_dir, binarizer.name)

        # Deserialize the Binarizer
        node_name = "{}.node".format(binarizer.name)
        binarizer_tf_ds = Binarizer()
        binarizer_tf_ds.deserialize_from_bundle(self.tmp_dir, node_name)

        # Transform some sample data
        res_a = binarizer.transform(self.df[['a']])
        res_b = binarizer_tf_ds.transform(self.df[['a']])

        self.assertEqual(res_a[0][0], res_b[0][0])
        self.assertEqual(res_a[1][0], res_b[1][0])
        self.assertEqual(res_a[2][0], res_b[2][0])
        self.assertEqual(res_a[3][0], res_b[3][0])

    def polynomial_expansion_test(self):

        extract_features = ['a']
        feature_extractor = FeatureExtractor(input_scalars=['a'],
                                         output_vector='extracted_a_output',
                                         output_vector_items=["{}_out".format(x) for x in extract_features])

        polynomial_exp = PolynomialFeatures(degree=2, include_bias=False)
        polynomial_exp.mlinit(prior_tf=feature_extractor,
                              output_features='poly')

        Xres = polynomial_exp.fit_transform(self.df[['a']])

        self.assertEqual(Xres[0][1], Xres[0][0] * Xres[0][0])

        polynomial_exp.serialize_to_bundle(self.tmp_dir, polynomial_exp.name)

        expected_model = {
          "op": "sklearn_polynomial_expansion",
          "attributes": {
              "combinations": {
                  "string": "[x0,x0^2]"
              }
          }
        }

        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, polynomial_exp.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual(expected_model['op'], model['op'])
        self.assertEqual(expected_model['attributes']['combinations']['string'], model['attributes']['combinations']['string'])

        # Test node.json
        with open("{}/{}.node/node.json".format(self.tmp_dir, polynomial_exp.name)) as json_data:
            node = json.load(json_data)

        self.assertEqual(polynomial_exp.name, node['name'])
        self.assertEqual(polynomial_exp.input_features, node['shape']['inputs'][0]['name'])
        self.assertEqual(polynomial_exp.output_features, node['shape']['outputs'][0]['name'])

    def math_unary_exp_test(self):

        math_unary_tf = MathUnary(input_features=['a'], output_features='exp_a', transform_type='exp')

        Xres = math_unary_tf.fit_transform(self.df.a)

        self.assertEqual(np.exp(self.df.a[0]), Xres[0])

        math_unary_tf.serialize_to_bundle(self.tmp_dir, math_unary_tf.name)

        expected_model = {
          "op": "math_unary",
          "attributes": {
            "operation": {
              "string": 'exp'
            }
          }
        }

        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, math_unary_tf.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual(expected_model['attributes']['operation']['string'], model['attributes']['operation']['string'])

        # Test node.json
        with open("{}/{}.node/node.json".format(self.tmp_dir, math_unary_tf.name)) as json_data:
            node = json.load(json_data)

        self.assertEqual(math_unary_tf.name, node['name'])
        self.assertEqual(math_unary_tf.input_features[0], node['shape']['inputs'][0]['name'])
        self.assertEqual(math_unary_tf.output_features, node['shape']['outputs'][0]['name'])

    def math_unary_deserialize_exp_test(self):

        math_unary_tf = MathUnary(input_features=['a'], output_features='exp_a', transform_type='exp')

        Xres = math_unary_tf.fit_transform(self.df.a)

        self.assertEqual(np.exp(self.df.a[0]), Xres[0])

        math_unary_tf.serialize_to_bundle(self.tmp_dir, math_unary_tf.name)

        node_name = "{}.node".format(math_unary_tf.name)
        math_unary_ds_tf = MathUnary()
        math_unary_ds_tf = math_unary_ds_tf.deserialize_from_bundle(self.tmp_dir, node_name)

        with open("{}/{}.node/model.json".format(self.tmp_dir, math_unary_tf.name)) as json_data:
            model = json.load(json_data)

        res_a = math_unary_tf.transform(self.df['a'])
        res_b = math_unary_ds_tf.transform(self.df['a'])

        self.assertEqual(res_a[0], res_b[0])

    def math_unary_sin_test(self):

        math_unary_tf = MathUnary(input_features=['a'], output_features='sin_a', transform_type='sin')

        Xres = math_unary_tf.fit_transform(self.df.a)

        self.assertEqual(np.sin(self.df.a[0]), Xres[0])

        math_unary_tf.serialize_to_bundle(self.tmp_dir, math_unary_tf.name)

        expected_model = {
          "op": "math_unary",
          "attributes": {
            "operation": {
              "string": 'sin'
            }
          }
        }

        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, math_unary_tf.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual(expected_model['attributes']['operation']['string'], model['attributes']['operation']['string'])

        # Test node.json
        with open("{}/{}.node/node.json".format(self.tmp_dir, math_unary_tf.name)) as json_data:
            node = json.load(json_data)

        self.assertEqual(math_unary_tf.name, node['name'])
        self.assertEqual(math_unary_tf.input_features[0], node['shape']['inputs'][0]['name'])
        self.assertEqual(math_unary_tf.output_features, node['shape']['outputs'][0]['name'])

    def math_binary_test(self):

        math_binary_tf = MathBinary(input_features=['a', 'b'], output_features='a_plus_b', transform_type='add')

        Xres = math_binary_tf.fit_transform(self.df[['a', 'b']])

        assert_frame_equal(pd.DataFrame(self.df.a + self.df.b, columns=['a_plus_b']), Xres)

        math_binary_tf.serialize_to_bundle(self.tmp_dir, math_binary_tf.name)

        expected_model = {
          "op": "math_binary",
          "attributes": {
            "operation": {
              "string": 'add'
            }
          }
        }

        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, math_binary_tf.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual(expected_model['attributes']['operation']['string'], model['attributes']['operation']['string'])

        # Test node.json
        with open("{}/{}.node/node.json".format(self.tmp_dir, math_binary_tf.name)) as json_data:
            node = json.load(json_data)

        self.assertEqual(math_binary_tf.name, node['name'])
        self.assertEqual(math_binary_tf.input_features[0], node['shape']['inputs'][0]['name'])
        self.assertEqual(math_binary_tf.input_features[1], node['shape']['inputs'][1]['name'])
        self.assertEqual(math_binary_tf.output_features, node['shape']['outputs'][0]['name'])

    def math_binary_deserialize_add_test(self):

        math_binary_tf = MathBinary(input_features=['a', 'b'], output_features='a_plus_b', transform_type='add')

        Xres = math_binary_tf.fit_transform(self.df[['a', 'b']])

        assert_frame_equal(pd.DataFrame(self.df.a + self.df.b, columns=['a_plus_b']), Xres)

        math_binary_tf.serialize_to_bundle(self.tmp_dir, math_binary_tf.name)

        node_name = "{}.node".format(math_binary_tf.name)
        math_binary_ds_tf = MathBinary()
        math_binary_ds_tf = math_binary_ds_tf.deserialize_from_bundle(self.tmp_dir, node_name)

        res_a = math_binary_tf.transform(self.df[['a', 'b']])
        res_b = math_binary_ds_tf.transform(self.df[['a', 'b']])

        # TODO: Deserialization on output_features has some issue. fix this.
        # assert_frame_equal(res_a, res_b)

    def math_binary_subtract_test(self):

        math_binary_tf = MathBinary(input_features=['a', 'b'], output_features='a_less_b', transform_type='sub')

        Xres = math_binary_tf.fit_transform(self.df[['a', 'b']])

        assert_frame_equal(pd.DataFrame(self.df.a - self.df.b, columns=['a_less_b']), Xres)

        math_binary_tf.serialize_to_bundle(self.tmp_dir, math_binary_tf.name)

        expected_model = {
          "op": "math_binary",
          "attributes": {
            "operation": {
              "string": 'sub'
            }
          }
        }

        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, math_binary_tf.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual(expected_model['attributes']['operation']['string'], model['attributes']['operation']['string'])

        # Test node.json
        with open("{}/{}.node/node.json".format(self.tmp_dir, math_binary_tf.name)) as json_data:
            node = json.load(json_data)

        self.assertEqual(math_binary_tf.name, node['name'])
        self.assertEqual(math_binary_tf.input_features[0], node['shape']['inputs'][0]['name'])
        self.assertEqual(math_binary_tf.input_features[1], node['shape']['inputs'][1]['name'])
        self.assertEqual(math_binary_tf.output_features, node['shape']['outputs'][0]['name'])

    def math_binary_multiply_test(self):

        math_binary_tf = MathBinary(input_features=['a', 'b'], output_features='a_mul_b', transform_type='mul')

        Xres = math_binary_tf.fit_transform(self.df[['a', 'b']])

        assert_frame_equal(pd.DataFrame(self.df.a * self.df.b, columns=['a_mul_b']), Xres)

        math_binary_tf.serialize_to_bundle(self.tmp_dir, math_binary_tf.name)

        expected_model = {
          "op": "math_binary",
          "attributes": {
            "operation": {
              "string": 'mul'
            }
          }
        }

        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, math_binary_tf.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual(expected_model['attributes']['operation']['string'], model['attributes']['operation']['string'])

        # Test node.json
        with open("{}/{}.node/node.json".format(self.tmp_dir, math_binary_tf.name)) as json_data:
            node = json.load(json_data)

        self.assertEqual(math_binary_tf.name, node['name'])
        self.assertEqual(math_binary_tf.input_features[0], node['shape']['inputs'][0]['name'])
        self.assertEqual(math_binary_tf.input_features[1], node['shape']['inputs'][1]['name'])
        self.assertEqual(math_binary_tf.output_features, node['shape']['outputs'][0]['name'])

    def math_binary_divide_test(self):

        math_binary_tf = MathBinary(input_features=['a', 'b'], output_features='a_div_b', transform_type='div')

        Xres = math_binary_tf.fit_transform(self.df[['a', 'b']])

        assert_frame_equal(pd.DataFrame(self.df.a / self.df.b, columns=['a_div_b']), Xres)

        math_binary_tf.serialize_to_bundle(self.tmp_dir, math_binary_tf.name)

        expected_model = {
          "op": "math_binary",
          "attributes": {
            "operation": {
              "string": 'div'
            }
          }
        }

        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, math_binary_tf.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual(expected_model['attributes']['operation']['string'], model['attributes']['operation']['string'])

        # Test node.json
        with open("{}/{}.node/node.json".format(self.tmp_dir, math_binary_tf.name)) as json_data:
            node = json.load(json_data)

        self.assertEqual(math_binary_tf.name, node['name'])
        self.assertEqual(math_binary_tf.input_features[0], node['shape']['inputs'][0]['name'])
        self.assertEqual(math_binary_tf.input_features[1], node['shape']['inputs'][1]['name'])
        self.assertEqual(math_binary_tf.output_features, node['shape']['outputs'][0]['name'])

    def string_map_test(self):

        df = pd.DataFrame(['test_one', 'test_two', 'test_one', 'test_one', 'test_two'], columns=['a'])
        string_map_tf = StringMap(input_features=['a'], output_features='a_mapped', labels={"test_one":1.0, "test_two": 0.0})

        Xres = string_map_tf.fit_transform(df)
        self.assertEqual(1.0, Xres[0])
        self.assertEqual(0.0, Xres[1])
        self.assertEqual(1.0, Xres[2])
        self.assertEqual(1.0, Xres[3])
        self.assertEqual(0.0, Xres[4])

        string_map_tf.serialize_to_bundle(self.tmp_dir, string_map_tf.name)

        expected_model = {
            "op": "string_map",
            "attributes": {
                "labels": {
                    "type": "list",
                    "string": ["test_one", "test_two"]
                },
                "values": {
                    "type": "list",
                    "double": [1.0, 0.0]
                }
            }
        }
        #
        # Test model.json
        with open("{}/{}.node/model.json".format(self.tmp_dir, string_map_tf.name)) as json_data:
            model = json.load(json_data)

        self.assertEqual(expected_model['attributes']['labels']['string'], model['attributes']['labels']['string'])
        self.assertEqual(expected_model['attributes']['values']['double'], model['attributes']['values']['double'])

        # Test node.json
        with open("{}/{}.node/node.json".format(self.tmp_dir, string_map_tf.name)) as json_data:
            node = json.load(json_data)

        self.assertEqual(string_map_tf.name, node['name'])
        self.assertEqual(string_map_tf.input_features[0], node['shape']['inputs'][0]['name'])
        self.assertEqual(string_map_tf.output_features, node['shape']['outputs'][0]['name'])

    def string_map_deserializer_test(self):

        df = pd.DataFrame(['test_one', 'test_two', 'test_one', 'test_one', 'test_two'], columns=['a'])
        string_map = StringMap(input_features=['a'], output_features='a_mapped', labels={"test_one":1.0, "test_two": 0.0})
        string_map.serialize_to_bundle(self.tmp_dir, string_map.name)

        # Now deserialize it back
        node_name = "{}.node".format(string_map.name)
        string_map_tf = StringMap()
        string_map_tf = string_map_tf.deserialize_from_bundle(self.tmp_dir, node_name)

        # Transform some sample data
        res_a = string_map.fit_transform(df)
        res_b = string_map_tf.fit_transform(df)

        self.assertEqual(res_a[0], res_b[0])
        self.assertEqual(res_a[1], res_b[1])
        self.assertEqual(res_a[2], res_b[2])
        self.assertEqual(res_a[3], res_b[3])
        self.assertEqual(res_a[4], res_b[4])
        self.assertEqual(string_map.name, string_map_tf.name)
        self.assertEqual(string_map.op, string_map_tf.op)
        self.assertEqual(string_map.labels, string_map_tf.labels)
